r"""
The submodule in `cl_algorithms` for [AdaHAT (Adaptive Hard Attention to the Task) algorithm](https://link.springer.com/chapter/10.1007/978-3-031-70352-2_9).
"""

__all__ = ["AdaHAT"]

import logging

import torch
from torch import Tensor

from clarena.backbones import HATMaskBackbone
from clarena.cl_algorithms import HAT
from clarena.cl_heads import HeadsCIL, HeadsTIL
from clarena.utils import HATNetworkCapacity

# always get logger for built-in logging in each module
pylogger = logging.getLogger(__name__)


class AdaHAT(HAT):
    r"""AdaHAT (Adaptive Hard Attention to the Task) algorithm.

    [Adaptive HAT (Adaptive Hard Attention to the Task, 2024)](https://link.springer.com/chapter/10.1007/978-3-031-70352-2_9) is an architecture-based continual learning approach that improves [HAT (Hard Attention to the Task, 2018)](http://proceedings.mlr.press/v80/serra18a) by introducing new adaptive soft gradient clipping based on parameter importance and network sparsity.

    We implement AdaHAT as a subclass of HAT algorithm, as AdaHAT has the same  `forward()`, `compensate_task_embedding_gradients()`, `training_step()`, `on_train_end()`,`validation_step()`, `test_step()` method as `HAT` class.
    """

    def __init__(
        self,
        backbone: HATMaskBackbone,
        heads: HeadsTIL | HeadsCIL,
        adjustment_mode: str,
        adjustment_intensity: float,
        s_max: float,
        clamp_threshold: float,
        mask_sparsity_reg_factor: float,
        mask_sparsity_reg_mode: str = "original",
        task_embedding_init_mode: str = "N01",
        epsilon: float = 0.1,
    ) -> None:
        r"""Initialise the AdaHAT algorithm with the network.

        **Args:**
        - **backbone** (`HATMaskBackbone`): must be a backbone network with HAT mask mechanism.
        - **heads** (`HeadsTIL` | `HeadsCIL`): output heads.
        - **adjustment_mode** (`str`): the strategy of adjustment i.e. the mode of gradient clipping, should be one of the following:
            1. 'adahat': set the gradients of parameters linking to masked units to a soft adjustment rate in the original AdaHAT approach. This is the way that AdaHAT does, which allowes the part of network for previous tasks to be updated slightly. See equation (8) and (9) chapter 3.1 in [AdaHAT paper](https://link.springer.com/chapter/10.1007/978-3-031-70352-2_9).
            2. 'adahat_no_sum': set the gradients of parameters linking to masked units to a soft adjustment rate in the original AdaHAT approach, but without considering the information of parameter importance i.e. summative mask. This is the way that one of the AdaHAT ablation study does. See chapter 4.3 in [AdaHAT paper](https://link.springer.com/chapter/10.1007/978-3-031-70352-2_9).
            3. 'adahat_no_reg': set the gradients of parameters linking to masked units to a soft adjustment rate in the original AdaHAT approach, but without considering the information of network sparsity i.e. mask sparsity regularisation value. This is the way that one of the AdaHAT ablation study does. See chapter 4.3 in [AdaHAT paper](https://link.springer.com/chapter/10.1007/978-3-031-70352-2_9).
        - **adjustment_intensity** (`float`): hyperparameter, control the overall intensity of gradient adjustment. It's the $\alpha$ in equation (9) in [AdaHAT paper](https://link.springer.com/chapter/10.1007/978-3-031-70352-2_9).
        - **s_max** (`float`): hyperparameter, the maximum scaling factor in the gate function. See chapter 2.4 "Hard Attention Training" in [HAT paper](http://proceedings.mlr.press/v80/serra18a).
        - **clamp_threshold** (`float`): the threshold for task embedding gradient compensation. See chapter 2.5 "Embedding Gradient Compensation" in [HAT paper](http://proceedings.mlr.press/v80/serra18a).
        - **mask_sparsity_reg_factor** (`float`): hyperparameter, the regularisation factor for mask sparsity.
        - **mask_sparsity_reg_mode** (`str`): the mode of mask sparsity regularisation, should be one of the following:
            1. 'original' (default): the original mask sparsity regularisation in HAT paper.
            2. 'cross': the cross version mask sparsity regularisation.
        - **task_embedding_init_mode** (`str`): the initialisation method for task embeddings, should be one of the following:
            1. 'N01' (default): standard normal distribution $N(0, 1)$.
            2. 'U-11': uniform distribution $U(-1, 1)$.
            3. 'U01': uniform distribution $U(0, 1)$.
            4. 'U-10': uniform distribution $U(-1, 0)$.
            5. 'last': inherit task embedding from last task.
        - **epsilon** (`float`): the value added to network sparsity to avoid division by zero appeared in equation (9) in [AdaHAT paper](https://link.springer.com/chapter/10.1007/978-3-031-70352-2_9).
        """
        HAT.__init__(
            self,
            backbone=backbone,
            heads=heads,
            adjustment_mode=adjustment_mode,
            s_max=s_max,
            clamp_threshold=clamp_threshold,
            mask_sparsity_reg_factor=mask_sparsity_reg_factor,
            mask_sparsity_reg_mode=mask_sparsity_reg_mode,
            task_embedding_init_mode=task_embedding_init_mode,
            alpha=None,
        )

        self.adjustment_intensity = adjustment_intensity
        r"""Store the adjustment intensity in equation (9) in [AdaHAT paper](https://link.springer.com/chapter/10.1007/978-3-031-70352-2_9)."""
        self.epsilon = epsilon
        """Store the small value to avoid division by zero appeared in equation (9) in [AdaHAT paper](https://link.springer.com/chapter/10.1007/978-3-031-70352-2_9)."""

        self.summative_mask_for_previous_tasks: dict[str, Tensor] = {}
        r"""Store the summative binary attention mask $\mathrm{M}^{<t,\text{sum}}$ previous tasks $1,\cdots, t-1$, gated from the task embedding. Keys are task IDs and values are the corresponding summative mask. Each cumulative mask is a dict where keys are layer names and values are the binary mask tensor for the layer. The mask tensor has size (number of units). """

        # set manual optimisation
        self.automatic_optimization = False

        AdaHAT.sanity_check(self)

    def sanity_check(self) -> None:
        r"""Check the sanity of the arguments.

        **Raises:**
        - **ValueError**: If the `adjustment_intensity` is not positive.
        """
        if self.adjustment_intensity <= 0:
            raise ValueError(
                f"The adjustment intensity should be positive, but got {self.adjustment_intensity}."
            )

    def on_train_start(self) -> None:
        r"""Additionally initialise the summative mask at the beginning of first task."""
        HAT.on_train_start(self)

        # initialise the summative mask at the beginning of first task. This should not be called in `__init__()` method as the `self.device` is not available at that time.
        if self.task_id == 1:
            for layer_name in self.backbone.weighted_layer_names:
                layer = self.backbone.get_layer_by_name(
                    layer_name
                )  # get the layer by its name
                num_units = layer.weight.shape[0]

                self.summative_mask_for_previous_tasks[layer_name] = torch.zeros(
                    num_units
                ).to(
                    self.device
                )  # the summative mask $\mathrm{M}^{<t,\text{sum}}$ is initialised as zeros mask ($t = 1$). See equation (7) in chapter 3.1 in [AdaHAT paper](https://link.springer.com/chapter/10.1007/978-3-031-70352-2_9).

    def clip_grad_by_adjustment(
        self,
        network_sparsity: dict[str, Tensor] | None = None,
    ) -> Tensor:
        r"""Clip the gradients by the adjustment rate.

        Note that as the task embedding fully covers every layer in the backbone network, no parameters are left out of this system. This applies not only the parameters in between layers with task embedding, but also those before the first layer. We designed it seperately in the codes.

        Network capacity is measured along with this method. Network capacity is defined as the average adjustment rate over all parameters. See chapter 4.1 in [AdaHAT paper](https://link.springer.com/chapter/10.1007/978-3-031-70352-2_9).

        **Args:**
        - **network_sparsity** (`dict[str, Tensor]` | `None`): The network sparsity i.e. the mask sparsity loss of each layer for the current task. It applies only to AdaHAT modes, as it is used to calculate the adjustment rate for the gradients.

        **Returns:**
        - **capacity** (`Tensor`): the calculated network capacity.
        """

        # initialise network capacity metric
        capacity = HATNetworkCapacity()

        # Calculate the adjustment rate for gradients of the parameters, both weights and biases (if exists)
        for layer_name in self.backbone.weighted_layer_names:

            layer = self.backbone.get_layer_by_name(
                layer_name
            )  # get the layer by its name

            # placeholder for the adjustment rate to avoid the error of using it before assignment
            adjustment_rate_weight = 1
            adjustment_rate_bias = 1

            weight_importance, bias_importance = (
                self.backbone.get_layer_measure_parameter_wise(
                    unit_wise_measure=self.summative_mask_for_previous_tasks,
                    layer_name=layer_name,
                    aggregation="min",
                )
            )  # AdaHAT depend on parameter importance instead of parameter mask like HAT

            network_sparsity_layer = network_sparsity[layer_name]

            if self.adjustment_mode == "adahat":
                r_layer = self.adjustment_intensity / (
                    self.epsilon + network_sparsity_layer
                )
                adjustment_rate_weight = torch.div(
                    r_layer, (weight_importance + r_layer)
                )
                adjustment_rate_bias = torch.div(r_layer, (bias_importance + r_layer))

            elif self.adjustment_mode == "adahat_no_sum":

                r_layer = self.adjustment_intensity / (
                    self.epsilon + network_sparsity_layer
                )
                adjustment_rate_weight = torch.div(r_layer, (self.task_id + r_layer))
                adjustment_rate_bias = torch.div(r_layer, (self.task_id + r_layer))

            elif self.adjustment_mode == "adahat_no_reg":

                r_layer = self.adjustment_intensity / (self.epsilon + 0.0)
                adjustment_rate_weight = torch.div(
                    r_layer, (weight_importance + r_layer)
                )
                adjustment_rate_bias = torch.div(r_layer, (bias_importance + r_layer))

            # apply the adjustment rate to the gradients
            layer.weight.grad.data *= adjustment_rate_weight
            if layer.bias is not None:
                layer.bias.grad.data *= adjustment_rate_bias

            # update network capacity metric
            capacity.update(adjustment_rate_weight, adjustment_rate_bias)

        return capacity.compute()

    def on_train_end(self) -> None:
        r"""Additionally update summative mask after training the task."""

        HAT.on_train_end(self)

        mask_t = self.masks[
            f"{self.task_id}"
        ]  # get stored mask for the current task again
        self.summative_mask_for_previous_tasks = {
            layer_name: self.summative_mask_for_previous_tasks[layer_name]
            + mask_t[layer_name]
            for layer_name in self.backbone.weighted_layer_names
        }



r"""
The submodule in `backbones` for CL backbone network bases.
"""

__all__ = ["CLBackbone", "HATMaskBackbone"]

import logging

import torch
from torch import Tensor, nn
from typing_extensions import override

# always get logger for built-in logging in each module
pylogger = logging.getLogger(__name__)


class CLBackbone(nn.Module):
    r"""The base class of continual learning backbone networks, inherited from `nn.Module`."""

    def __init__(self, output_dim: int | None) -> None:
        r"""Initialise the CL backbone network.

        **Args:**
        - **output_dim** (`int` | `None`): The output dimension which connects to CL output heads. The `input_dim` of output heads are expected to be the same as this `output_dim`. In some cases, this class is used for a block in the backbone network, which doesn't have the output dimension. In this case, it can be `None`.
        """
        nn.Module.__init__(self)

        self.output_dim: int = output_dim
        r"""Store the output dimension of the backbone network."""

        self.weighted_layer_names: list[str] = []
        r"""Maintain a list of the weighted layer names. Weighted layer has weights connecting to other weighted layer. They are the main part of neural networks. **It must be provided in subclasses.**
        
        The names are following the `nn.Module` internal naming mechanism. For example, if the a layer is assigned to `self.conv1`, the name becomes `conv1`. If the `nn.Sequential` is used, the name becomes the index of the layer in the sequence, such as `0`, `1`, etc. If hierarchical structure is used, for example, a `nn.Module` is assigned to `self.block` which has `self.conv1`, the name becomes `block/conv1`. Note that it should be `block.conv1` according to `nn.Module` internal mechanism, but we use '/' instead of '.' to avoid the error of using '.' in the key of `ModuleDict`.
        
        In HAT architecture, it's also the layer names with task embedding masking in the order of forward pass. HAT gives task embedding to every possible weighted layer. 
        """

        self.task_id: int
        r"""Task ID counter indicating which task is being processed. Self updated during the task loop."""

    def setup_task_id(self, task_id: int) -> None:
        r"""Set up which task's dataset the CL experiment is on. This must be done before `forward()` method is called.

        **Args:**
        - **task_id** (`int`): the target task ID.
        """
        self.task_id = task_id

    def get_layer_by_name(self, layer_name: str) -> nn.Module:
        r"""Get the layer by its name.

        **Args:**
        - **layer_name** (`str`): the name of the layer. Note that the name is the name substituting the '.' with '/', like `block/conv1`, rather than `block.conv1`.

        **Returns:**
        - **layer** (`nn.Module`): the layer.
        """
        for name, layer in self.named_modules():
            if name == layer_name.replace("/", "."):
                return layer

    def preceding_layer_name(self, layer_name: str) -> str:
        r"""Get the name of the preceding layer of the given layer from the stored `self.masked_layer_order`. If the given layer is the first layer, return `None`.

        **Args:**
        - **layer_name** (`str`): the name of the layer.

        **Returns:**
        - **preceding_layer_name** (`str`): the name of the preceding layer.

        **Raises:**
        - **ValueError**: if `layer_name` is not in the weighted layer order.
        """

        if layer_name not in self.weighted_layer_names:
            raise ValueError(f"The layer name {layer_name} doesn't exist.")

        weighted_layer_idx = self.weighted_layer_names.index(layer_name)
        if weighted_layer_idx == 0:
            return None
        return self.weighted_layer_names[weighted_layer_idx - 1]

    @override  # since `nn.Module` uses it
    def forward(
        self,
        input: Tensor,
        stage: str,
        task_id: int | None = None,
    ) -> tuple[Tensor, dict[str, Tensor]]:
        r"""The forward pass for data from task `task_id`. In some backbones, the forward pass might be different for different tasks. **It must be implemented by subclasses.**

        **Args:**
        - **input** (`Tensor`): The input tensor from data.
        - **stage** (`str`): the stage of the forward pass, should be one of the following:
            1. 'train': training stage.
            2. 'validation': validation stage.
            3. 'test': testing stage.
        - **task_id** (`int` | `None`): the task ID where the data are from. If stage is 'train' or 'validation', it is usually from the current task `self.task_id`. If stage is 'test', it could be from any seen task. In TIL, the task IDs of test data are provided thus this argument can be used. In CIL, they are not provided, so it is just a placeholder for API consistence but never used, and best practices are not to provide this argument and leave it as the default value.

        **Returns:**
        - **output_feature** (`Tensor`): the output feature tensor to be passed into heads. This is the main target of backpropagation.
        - **hidden_features** (`dict[str, Tensor]`): the hidden features (after activation) in each weighted layer. Key (`str`) is the weighted layer name, value (`Tensor`) is the hidden feature tensor. This is used for the continual learning algorithms that need to use the hidden features for various purposes.
        """


class HATMaskBackbone(CLBackbone):
    r"""The backbone network for HAT-based algorithms with learnable hard attention masks.

    HAT-based algorithms:

    - [**HAT (Hard Attention to the Task, 2018)**](http://proceedings.mlr.press/v80/serra18a) is an architecture-based continual learning approach that uses learnable hard attention masks to select the task-specific parameters.
    - [**Adaptive HAT (Adaptive Hard Attention to the Task, 2024)**](https://link.springer.com/chapter/10.1007/978-3-031-70352-2_9) is an architecture-based continual learning approach that improves [HAT (Hard Attention to the Task, 2018)](http://proceedings.mlr.press/v80/serra18a) by introducing new adaptive soft gradient clipping based on parameter importance and network sparsity.
    - **CBPHAT** is what I am working on, trying combining HAT (Hard Attention to the Task) algorithm with Continual Backpropagation (CBP) by leveraging the contribution utility as the parameter importance like in AdaHAT (Adaptive Hard Attention to the Task) algorithm.
    """

    def __init__(self, output_dim: int | None, gate: str) -> None:
        r"""Initialise the HAT mask backbone network with task embeddings and masks.

        **Args:**
        - **output_dim** (`int`): The output dimension which connects to CL output heads. The `input_dim` of output heads are expected to be the same as this `output_dim`. In some cases, this class is used for a block in the backbone network, which doesn't have the output dimension. In this case, it can be `None`.
        - **gate** (`str`): the type of gate function turning the real value task embeddings into attention masks, should be one of the following:
            - `sigmoid`: the sigmoid function.
        """
        CLBackbone.__init__(self, output_dim=output_dim)

        self.register_hat_mask_module_explicitly(
            gate=gate
        )  # we moved the registration of the modules to a separate method to solve a problem of multiple inheritance in terms of `nn.Module`

        HATMaskBackbone.sanity_check(self)

    def register_hat_mask_module_explicitly(self, gate: str) -> None:
        r"""Register all `nn.Module`s explicitly in this method. For `HATMaskBackbone`, they are task embedding for the current task and the masks.

        **Args:**
        - **gate** (`str`): the type of gate function turning the real value task embeddings into attention masks, should be one of the following:
            - `sigmoid`: the sigmoid function.
        """
        self.gate: str = gate
        r"""Store the type of gate function."""
        if gate == "sigmoid":
            self.gate_fn: nn.Module = nn.Sigmoid()
            r"""The gate function turning the real value task embeddings into attention masks."""

        self.task_embedding_t: nn.ModuleDict = nn.ModuleDict()
        r"""Store the task embedding for the current task. Keys are the layer names and values are the task embedding `nn.Embedding` for the layer. Each task embedding has size (1, number of units).
        
        We use `ModuleDict` rather than `dict` to make sure `LightningModule` can properly register these model parameters for the purpose of, like automatically transfering to device, being recorded in model summaries.
        
        we use `nn.Embedding` rather than `nn.Parameter` to store the task embedding for each layer, which is a type of `nn.Module` and can be accepted by `nn.ModuleDict`. (`nn.Parameter` cannot be accepted by `nn.ModuleDict`.)
        
        **This must be defined to cover each weighted layer (just as `self.weighted_layer_names` listed) in the backbone network.** Otherwise, the uncovered parts will keep updating for all tasks and become a source of catastrophic forgetting. """

    def initialise_task_embedding(self, mode: str) -> None:
        r"""Initialise the task embedding for the current task.

        **Args:**
        - **mode** (`str`): the initialisation mode for task embeddings, should be one of the following:
            1. 'N01' (default): standard normal distribution $N(0, 1)$.
            2. 'U-11': uniform distribution $U(-1, 1)$.
            3. 'U01': uniform distribution $U(0, 1)$.
            4. 'U-10': uniform distribution $U(-1, 0)$.
            5. 'last': inherit task embedding from last task.
        """
        for te in self.task_embedding_t.values():
            if mode == "N01":
                nn.init.normal_(te.weight, 0, 1)
            elif mode == "U-11":
                nn.init.uniform_(te.weight, -1, 1)
            elif mode == "U01":
                nn.init.uniform_(te.weight, 0, 1)
            elif mode == "U-10":
                nn.init.uniform_(te.weight, -1, 0)
            elif mode == "last":
                pass

    def sanity_check(self) -> None:
        r"""Check the sanity of the arguments.

        **Raises:**
        - **ValueError**: when the `gate` is not one of the valid options.
        """

        if self.gate not in ["sigmoid"]:
            raise ValueError("The gate should be one of 'sigmoid'.")

    def get_mask(
        self,
        stage: str,
        s_max: float | None = None,
        batch_idx: int | None = None,
        num_batches: int | None = None,
        test_mask: dict[str, Tensor] | None = None,
    ) -> dict[str, Tensor]:
        r"""Get the hard attention mask used in `forward()` method for different stages.

        **Args:**
        - **stage** (`str`): the stage when applying the conversion, should be one of the following:
            1. 'train': training stage. If stage is 'train', get the mask from task embedding of current task through the gate function, which is scaled by an annealed scalar. See chapter 2.4 "Hard Attention Training" in [HAT paper](http://proceedings.mlr.press/v80/serra18a).
            2. ‘validation': validation stage. If stage is 'validation', get the mask from task embedding of current task through the gate function, which is scaled by `s_max`. (Note that in this stage, the binary mask hasn't been stored yet as the training is not over.)
            3. 'test': testing stage. If stage is 'test', apply the mask gate function is scaled by `s_max`, the large scaling making masks nearly binary.
        - **s_max** (`float`): the maximum scaling factor in the gate function. Doesn't apply to testing stage. See chapter 2.4 "Hard Attention Training" in [HAT paper](http://proceedings.mlr.press/v80/serra18a).
        - **batch_idx** (`int` | `None`): the current batch index. Applies only to training stage. For other stages, it is default `None`.
        - **num_batches** (`int` | `None`): the total number of batches. Applies only to training stage. For other stages, it is default `None`.
        - **test_mask** (`dict[str, Tensor]` | `None`): the binary mask used for test. Applies only to testing stage. For other stages, it is default `None`.

        **Returns:**
        - **mask** (`dict[str, Tensor]`): the hard attention (whose values are 0 or 1) mask. Key (`str`) is layer name, value (`Tensor`) is the mask tensor. The mask tensor has size (number of units).

        **Raises:**
        - **ValueError**: if the `batch_idx` and `batch_num` are not provided in 'train' stage; if the `s_max` is not provided in 'validation' stage; if the `task_id` is not provided in 'test' stage.
        """

        # sanity check
        if stage == "train" and (
            s_max is None or batch_idx is None or num_batches is None
        ):
            raise ValueError(
                "The `s_max`, `batch_idx` and `batch_num` should be provided at training stage, instead of the default value `None`."
            )
        if stage == "validation" and (s_max is None):
            raise ValueError(
                "The `s_max` should be provided at validation stage, instead of the default value `None`."
            )
        if stage == "test" and (test_mask is None):
            raise ValueError(
                "The `task_mask` should be provided at testing stage, instead of the default value `None`."
            )

        mask = {}
        if stage == "train":
            for layer_name in self.weighted_layer_names:
                anneal_scalar = 1 / s_max + (s_max - 1 / s_max) * (batch_idx - 1) / (
                    num_batches - 1
                )  # see equation (3) in chapter 2.4 "Hard Attention Training" in [HAT paper](http://proceedings.mlr.press/v80/serra18a).
                mask[layer_name] = self.gate_fn(
                    self.task_embedding_t[layer_name].weight * anneal_scalar
                ).squeeze()
        elif stage == "validation":
            for layer_name in self.weighted_layer_names:
                mask[layer_name] = self.gate_fn(
                    self.task_embedding_t[layer_name].weight * s_max
                ).squeeze()
        elif stage == "test":
            mask = test_mask

        return mask

    def get_cumulative_mask(self) -> dict[str, Tensor]:
        r"""Get the cumulative mask till current task.

        **Returns:**
        - **cumulative_mask** (`dict[str, Tensor]`): the cumulative mask. Key (`str`) is layer name, value (`Tensor`) is the mask tensor. The mask tensor has size (number of units).
        """
        return self.cumulative_mask_for_previous_tasks

    def get_summative_mask(self) -> dict[str, Tensor]:
        r"""Get the summative mask till current task.

        **Returns:**
        - **summative_mask** (`dict[str, Tensor]`): the summative mask tensor. Key (`str`) is layer name, value (`Tensor`) is the mask tensor. The mask tensor has size (number of units).
        """
        return self.summative_mask_for_previous_tasks

    def get_layer_measure_parameter_wise(
        self,
        unit_wise_measure: dict[str, Tensor],
        layer_name: str,
        aggregation: str,
    ) -> Tensor:
        r"""Get the parameter-wise measure on the parameters right before the given layer.

        It is calculated from the given unit-wise measure. It aggregates two feature-sized vectors (corresponding the given layer and preceding layer) into a weight-wise matrix (corresponding the weights in between) and bias-wise vector (corresponding the bias of the given layer), using the given aggregation method. For example, given two feature-sized measure $m_{l,i}$ and $m_{l-1,j}$ and 'min' aggregation, the parameter-wise measure is then $\min \left(a_{l,i}, a_{l-1,j}\right)$, a matrix with respect to $i, j$.

        Note that if the given layer is the first layer with no preceding layer, we will get parameter-wise measure directly broadcasted from the unit-wise measure of given layer.

        This method is used in the calculation of parameter-wise measure in various HAT-based algorithms:

        - **HAT**: the parameter-wise measure is the binary mask for previous tasks from the unit-wise cumulative mask of previous tasks `self.cumulative_mask_for_previous_tasks`, which is $\min \left(a_{l,i}^{<t}, a_{l-1,j}^{<t}\right)$ in equation (2) in [HAT paper](http://proceedings.mlr.press/v80/serra18a).
        - **AdaHAT**: the parameter-wise measure is the parameter importance for previous tasks from the unit-wise summative mask of previous tasks `self.summative_mask_for_previous_tasks`, which is $\min \left(m_{l,i}^{<t,\text{sum}}, m_{l-1,j}^{<t,\text{sum}}\right)$ in equation (9) in [AdaHAT paper](https://link.springer.com/chapter/10.1007/978-3-031-70352-2_9).
        - **CBPHAT**: the parameter-wise measure is the parameter importance for previous tasks from the unit-wise importance of previous tasks `self.unit_importance_for_previous_tasks` based on contribution utility, which is $\min \left(I_{l,i}^{(t-1)}, I_{l-1,j}^{(t-1)}\right)$ in the adjustment rate formula in the paper draft.

        **Args:**
        - **unit_wise_measure** (`dict[str, Tensor]`): the unit-wise measure. Key is layer name, value is the unit-wise measure tensor. The measure tensor has size (number of units).
        - **layer_name** (`str`): the name of given layer.
        - **aggregation** (`str`): the aggregation method turning two feature-wise measures into weight-wise matrix, should be one of the following:
            - 'min': takes minimum of the two connected unit measures.
            - 'max': takes maximum of the two connected unit measures.

        **Returns:**
        - **weight_measure** (`Tensor`): the weight measure matrix, same size as the corresponding weights.
        - **bias_measure** (`Tensor`): the bias measure vector, same size as the corresponding bias.


        """

        # initialise the aggregation function
        if aggregation == "min":
            aggregation_func = torch.min
        elif aggregation == "max":
            aggregation_func = torch.max
        else:
            raise ValueError(f"The aggregation method {aggregation} is not supported.")

        # get the preceding layer name
        preceding_layer_name = self.preceding_layer_name(layer_name)

        # get weight size for expanding the measures
        layer = self.get_layer_by_name(layer_name)
        weight_size = layer.weight.size()

        # construct the weight-wise measure
        layer_measure = unit_wise_measure[layer_name]
        layer_measure_broadcast_size = (-1, 1) + tuple(
            1 for _ in range(len(weight_size) - 2)
        )  # since the size of mask tensor is (number of units), we extend it to (number of units, 1) and expand it to the weight size. The weight size has 2 dimensions in fully connected layers and 4 dimensions in convolutional layers

        layer_measure_broadcasted = layer_measure.view(
            *layer_measure_broadcast_size
        ).expand(
            weight_size,
        )  # expand the given layer mask to the weight size and broadcast

        if (
            preceding_layer_name
        ):  # if the layer is not the first layer, where the preceding layer exists

            preceding_layer_measure_broadcast_size = (1, -1) + tuple(
                1 for _ in range(len(weight_size) - 2)
            )  # since the size of mask tensor is (number of units), we extend it to (1, number of units) and expand it to the weight size. The weight size has 2 dimensions in fully connected layers and 4 dimensions in convolutional layers
            preceding_layer_measure = unit_wise_measure[preceding_layer_name]
            preceding_layer_measure_broadcasted = preceding_layer_measure.view(
                *preceding_layer_measure_broadcast_size
            ).expand(
                weight_size
            )  # expand the preceding layer mask to the weight size and broadcast
            weight_measure = aggregation_func(
                layer_measure_broadcasted, preceding_layer_measure_broadcasted
            )  # get the minimum of the two mask vectors, from expanded
        else:  # if the layer is the first layer
            weight_measure = layer_measure_broadcasted

        # construct the bias-wise measure
        bias_measure = layer_measure

        return weight_measure, bias_measure

    @override
    def forward(
        self,
        input: Tensor,
        stage: str,
        s_max: float | None = None,
        batch_idx: int | None = None,
        num_batches: int | None = None,
        test_mask: dict[str, Tensor] | None = None,
    ) -> tuple[Tensor, dict[str, Tensor], dict[str, Tensor]]:
        r"""The forward pass for data from task `task_id`. Task-specific mask for `task_id` are applied to the units in each layer.

        **Args:**
        - **input** (`Tensor`): The input tensor from data.
        - **stage** (`str`): the stage of the forward pass, should be one of the following:
            1. 'train': training stage.
            2. 'validation': validation stage.
            3. 'test': testing stage.
        - **s_max** (`float`): the maximum scaling factor in the gate function. See chapter 2.4 "Hard Attention Training" in [HAT paper](http://proceedings.mlr.press/v80/serra18a).
        - **batch_idx** (`int` | `None`): the current batch index. Applies only to training stage. For other stages, it is default `None`.
        - **num_batches** (`int` | `None`): the total number of batches. Applies only to training stage. For other stages, it is default `None`.
        - **test_mask** (`dict[str, Tensor]` | `None`): the binary mask used for test. Applies only to testing stage. For other stages, it is default `None`.

        **Returns:**
        - **output_feature** (`Tensor`): the output feature tensor to be passed into heads. This is the main target of backpropagation.
        - **mask** (`dict[str, Tensor]`): the mask for the current task. Key (`str`) is layer name, value (`Tensor`) is the mask tensor. The mask tensor has size (number of units).
        - **hidden_features** (`dict[str, Tensor]`): the hidden features (after activation) in each weighted layer. Key (`str`) is the weighted layer name, value (`Tensor`) is the hidden feature tensor. This is used for the continual learning algorithms that need to use the hidden features for various purposes. Although HAT algorithm does not need this, it is still provided for API consistence for other HAT-based algorithms inherited this `forward()` method of `HAT` class.

        """
        # this should be copied to all subclasses. Make sure it is called to get the mask for the current task from the task embedding in this stage
        mask = self.get_mask(
            stage,
            s_max=s_max,
            batch_idx=batch_idx,
            num_batches=num_batches,
            test_mask=test_mask,
        )



r"""
The submodule in `cl_algorithms` for CBPHAT algorithm.
"""

__all__ = ["CBPHAT"]

import logging
from typing import Any

import torch
from torch import Tensor

from clarena.backbones import HATMaskBackbone
from clarena.cl_algorithms import AdaHAT
from clarena.cl_heads import HeadsCIL, HeadsTIL
from clarena.utils import HATNetworkCapacity
from clarena.utils.transforms import min_max_normalise

# always get logger for built-in logging in each module
pylogger = logging.getLogger(__name__)


class CBPHAT(AdaHAT):
    r"""CBPHAT algorithm.

    CBPHAT is what I am working on, trying combining HAT (Hard Attention to the Task) algorithm with Continual Backpropagation (CBP) by leveraging the contribution utility as the parameter importance like in AdaHAT (Adaptive Hard Attention to the Task) algorithm.

    We implement CBPHAT as a subclass of AdaHAT algorithm because CBPHAT adopt the similar idea as AdaHAT.
    """

    def __init__(
        self,
        backbone: HATMaskBackbone,
        heads: HeadsTIL | HeadsCIL,
        adjustment_mode: str,
        adjustment_intensity: float,
        utility_decay_rate: float,
        s_max: float,
        clamp_threshold: float,
        mask_sparsity_reg_factor: float,
        mask_sparsity_reg_mode: str = "original",
        task_embedding_init_mode: str = "N01",
        epsilon: float = 0.1,
    ) -> None:
        r"""Initialise the CBPHAT algorithm with the network.

        **Args:**
        - **backbone** (`HATMaskBackbone`): must be a backbone network with HAT mask mechanism.
        - **heads** (`HeadsTIL` | `HeadsCIL`): output heads.
        - **adjustment_mode** (`str`): the strategy of adjustment i.e. the mode of gradient clipping, should be one of the following:
            1. 'cbphat': our original CBP mode.
        - **adjustment_intensity** (`float`): hyperparameter, control the overall intensity of gradient adjustment. It's the $\alpha$ in equation (9) in [AdaHAT paper](https://link.springer.com/chapter/10.1007/978-3-031-70352-2_9).
        - **utility_decay_rate** (`float`): the utility decay rate of units. It is the rate at which the utility of a unit decays over time.
        - **s_max** (`float`): hyperparameter, the maximum scaling factor in the gate function. See chapter 2.4 "Hard Attention Training" in [HAT paper](http://proceedings.mlr.press/v80/serra18a).
        - **clamp_threshold** (`float`): the threshold for task embedding gradient compensation. See chapter 2.5 "Embedding Gradient Compensation" in [HAT paper](http://proceedings.mlr.press/v80/serra18a).
        - **mask_sparsity_reg_factor** (`float`): hyperparameter, the regularisation factor for mask sparsity.
        - **mask_sparsity_reg_mode** (`str`): the mode of mask sparsity regularisation, should be one of the following:
            1. 'original' (default): the original mask sparsity regularisation in HAT paper.
            2. 'cross': the cross version mask sparsity regularisation.
        - **task_embedding_init_mode** (`str`): the initialisation method for task embeddings, should be one of the following:
            1. 'N01' (default): standard normal distribution $N(0, 1)$.
            2. 'U-11': uniform distribution $U(-1, 1)$.
            3. 'U01': uniform distribution $U(0, 1)$.
            4. 'U-10': uniform distribution $U(-1, 0)$.
            5. 'last': inherit task embedding from last task.
        - **epsilon** (`float`): the value added to network sparsity to avoid zero appeared in equation (9) in [AdaHAT paper](https://link.springer.com/chapter/10.1007/978-3-031-70352-2_9).
        """
        AdaHAT.__init__(
            self,
            backbone=backbone,
            heads=heads,
            adjustment_mode=adjustment_mode,
            adjustment_intensity=adjustment_intensity,
            s_max=s_max,
            clamp_threshold=clamp_threshold,
            mask_sparsity_reg_factor=mask_sparsity_reg_factor,
            mask_sparsity_reg_mode=mask_sparsity_reg_mode,
            task_embedding_init_mode=task_embedding_init_mode,
            epsilon=epsilon,
        )

        self.utility_decay_rate: float = utility_decay_rate
        r"""Store the utility decay rate of units. """

        self.contribution_utility_t: dict[str, Tensor] = {}
        r"""Store the summative min-max scaled contribution utility of units. See $U$ in the paper draft. Keys are layer names and values are the utility tensor for the layer. The utility tensor is the same size as the feature tensor with size (number of units). """
        self.unit_importance_for_previous_tasks: dict[str, Tensor] = {}
        r"""Store the unit importance values of units for previous tasks (1, \cdots, self.task_id - 1). See the "screenshot" $I^{(t-1)}$ in the paper draft. Keys are layer names and values are the importance tensor for the layer. The importance tensor is the same size as the feature tensor with size (number of units). """
        self.age_t: dict[str, Tensor] = {}
        r"""Store the age of units. Keys are layer names and values are the age tensor for the layer for current task. The age tensor is the same size as the feature tensor with size (number of units). """

        # set manual optimisation
        self.automatic_optimization = False

        CBPHAT.sanity_check(self)

    def sanity_check(self) -> None:
        r"""Check the sanity of the arguments.

        **Raises:**
        - **ValueError**: If the utility decay rate is not in the range (0, 1].
        """
        if self.utility_decay_rate > 1 or self.utility_decay_rate <= 0:
            raise ValueError(
                f"The utility decay rate should be in the range (0, 1], but got {self.utility_decay_rate}."
            )

    def on_train_start(self) -> None:
        r"""Additionally initialise the utility, age and the CBPHAT unit importance for each layer as zeros."""
        AdaHAT.on_train_start(self)

        for layer_name in self.backbone.weighted_layer_names:
            layer = self.backbone.get_layer_by_name(
                layer_name
            )  # get the layer by its name
            num_units = layer.weight.shape[0]

            # initialise the utility and age at the beginning of first task
            self.contribution_utility_t[layer_name] = torch.zeros(num_units).to(
                self.device
            )
            self.age_t[layer_name] = torch.zeros(num_units).to(self.device)

            # initialise the unit importance at the beginning of first task. This should not be called in `__init__()` method as the `self.device` is not available at that time.
            if self.task_id == 1:
                self.unit_importance_for_previous_tasks[layer_name] = torch.zeros(
                    num_units
                ).to(
                    self.device
                )  # the unit importance $I^{(t-1)}$ is initialised as zeros mask ($t = 1$). See the paper draft.

    def clip_grad_by_adjustment(
        self,
        network_sparsity: dict[str, Tensor] | None = None,
    ) -> Tensor:
        r"""Clip the gradients by the adjustment rate.

        Note that as the task embedding fully covers every layer in the backbone network, no parameters are left out of this system. This applies not only the parameters in between layers with task embedding, but also those before the first layer. We designed it seperately in the codes.

        Network capacity is measured along with this method. Network capacity is defined as the average adjustment rate over all parameters. See chapter 4.1 in [AdaHAT paper](https://link.springer.com/chapter/10.1007/978-3-031-70352-2_9).


        **Args:**
        - **network_sparsity** (`dict[str, Tensor]` | `None`): The network sparsity i.e. the mask sparsity loss of each layer for the current task. It applies only to AdaHAT modes, as it is used to calculate the adjustment rate for the gradients.

        **Returns:**
        - **capacity** (`Tensor`): the calculated network capacity.
        """

        # initialise network capacity metric
        capacity = HATNetworkCapacity()

        # Calculate the adjustment rate for gradients of the parameters, both weights and biases (if exists)
        for layer_name in self.backbone.weighted_layer_names:

            layer = self.backbone.get_layer_by_name(
                layer_name
            )  # get the layer by its name

            # placeholder for the adjustment rate to avoid the error of using it before assignment
            adjustment_rate_weight = 1
            adjustment_rate_bias = 1

            weight_importance, bias_importance = (
                self.backbone.get_layer_measure_parameter_wise(
                    unit_wise_measure=self.unit_importance_for_previous_tasks,
                    layer_name=layer_name,
                    aggregation="min",
                )
            )

            weight_summative_mask, bias_summative_mask = (
                self.backbone.get_layer_measure_parameter_wise(
                    unit_wise_measure=self.summative_mask_for_previous_tasks,
                    layer_name=layer_name,
                    aggregation="min",
                )
            )

            network_sparsity_layer = network_sparsity[layer_name]

            if self.adjustment_mode == "cbphat":

                r_layer = self.adjustment_intensity / (
                    self.epsilon + network_sparsity_layer
                )
                adjustment_rate_weight = torch.div(
                    r_layer, (weight_importance + weight_summative_mask + r_layer)
                )

                adjustment_rate_bias = torch.div(
                    r_layer, (bias_importance + bias_summative_mask + r_layer)
                )

            # apply the adjustment rate to the gradients
            layer.weight.grad.data *= adjustment_rate_weight
            if layer.bias is not None:
                layer.bias.grad.data *= adjustment_rate_bias

            # update network capacity metric
            capacity.update(adjustment_rate_weight, adjustment_rate_bias)

        return capacity.compute()

    def on_train_batch_end(
        self, outputs: dict[str, Any], batch: Any, batch_idx: int
    ) -> None:
        r"""Update the contribution utility and age of units after each training step.

        **Args:**
        - **outputs** (`dict[str, Any]`): the outputs of the training step, which is the returns of the `training_step()` method in the `CLAlgorithm`.
        - **batch** (`Any`): the training data batch.
        - **batch_idx** (`int`): the index of the current batch. This is for the file name of mask figures.
        """

        hidden_features = outputs["hidden_features"]
        mask = outputs["mask"]

        for layer_name in self.backbone.weighted_layer_names:
            # layer-wise operation
            layer = self.backbone.get_layer_by_name(
                layer_name
            )  # get the layer by its name

            # update age
            self.age_t[layer_name] += 1

            # calculate current contribution utility
            current_contribution_utility = (
                (
                    torch.mean(
                        torch.abs(hidden_features[layer_name]),
                        dim=[
                            i
                            for i in range(hidden_features[layer_name].dim())
                            if i != 1
                        ],  # average the features over batch samples
                    )
                    * torch.sum(
                        torch.abs(layer.weight),
                        dim=[
                            i for i in range(layer.weight.dim()) if i != 0
                        ],  # sum over the output dimension
                    )
                )
                * mask[layer_name]
            ).detach()
            current_contribution_utility = min_max_normalise(
                current_contribution_utility
            )  # normalise the utility to [0,1] to avoid linearly increasing utility

            # update utility
            self.contribution_utility_t[layer_name] = (
                self.utility_decay_rate * self.contribution_utility_t[layer_name]
                + current_contribution_utility
            )

    def on_train_end(self) -> None:
        r"""Additionally convert the contribution utility into importance and store (take screenshot of) it as unit importance for previous tasks at the end of a task training."""
        AdaHAT.on_train_end(
            self
        )  # store the mask and update cumulative and summative masks

        for layer_name in self.backbone.weighted_layer_names:
            self.unit_importance_for_previous_tasks[layer_name] += (
                self.contribution_utility_t[layer_name]
            ) / self.age_t[layer_name]



r"""
The submodule in `cl_datasets` for Split CIFAR-100 dataset.
"""

__all__ = ["SplitCIFAR100"]

import logging
from typing import Callable

from torch.utils.data import Dataset, random_split
from torchvision.datasets import CIFAR100
from torchvision.transforms import transforms

from clarena.cl_datasets import CLSplitDataset

# always get logger for built-in logging in each module
pylogger = logging.getLogger(__name__)


class SplitCIFAR100(CLSplitDataset):
    r"""Split CIFAR-100 dataset. The [original CIFAR100 dataset](https://www.cs.toronto.edu/~kriz/cifar.html) is a subset of the 80 million tiny images dataset. It consists of 60,000 32x32 colour images in 100 classes, with 600 images per class. There are 50,000 training examples and 10,000 test examples."""

    num_classes: int = 100
    r"""The number of classes in CIFAR-100 dataset."""

    mean_original: tuple[float] = (0.5074, 0.4867, 0.4411)
    r"""The mean values for normalisation."""

    std_original: tuple[float] = (0.2011, 0.1987, 0.2025)
    r"""The standard deviation values for normalisation."""

    def __init__(
        self,
        root: str,
        num_tasks: int,
        class_split: list[list[int]],
        validation_percentage: float,
        batch_size: int = 1,
        num_workers: int = 8,
        custom_transforms: Callable | transforms.Compose | None = None,
        custom_target_transforms: Callable | transforms.Compose | None = None,
    ) -> None:
        r"""Initialise the Split CIFAR-100 dataset.

        **Args:**
        - **root** (`str`): the root directory where the original CIFAR-100 data 'cifar-100-python/' live.
        - **num_tasks** (`int`): the maximum number of tasks supported by the CL dataset.
        - **class_split** (`list[list[int]]`): the class split for each task. Each element in the list is a list of class labels (integers starting from 0) to split for a task.
        - **validation_percentage** (`float`): the percentage to randomly split some of the training data into validation data.
        - **batch_size** (`int`): The batch size in train, val, test dataloader.
        - **num_workers** (`int`): the number of workers for dataloaders.
        - **custom_transforms** (`transform` or `transforms.Compose` or `None`): the custom transforms to apply to ONLY TRAIN dataset. Can be a single transform, composed transforms or no transform.
        `ToTensor()`, normalise, permute and so on are not included.
        - **custom_target_transforms** (`transform` or `transforms.Compose` or `None`): the custom target transforms to apply to dataset labels. Can be a single transform, composed transforms or no transform. CL class mapping is not included.
        - **permutation_mode** (`str`): the mode of permutation, should be one of the following:
            1. 'all': permute all pixels.
            2. 'by_channel': permute channel by channel separately. All channels are applied the same permutation order.
            3. 'first_channel_only': permute only the first channel.
        - **permutation_seeds** (`list[int]` or `None`): the seeds for permutation operations used to construct tasks. Make sure it has the same number of seeds as `num_tasks`. Default is None, which creates a list of seeds from 1 to `num_tasks`.
        """
        CLSplitDataset.__init__(
            self,
            root=root,
            num_tasks=num_tasks,
            class_split=class_split,
            validation_percentage=validation_percentage,
            batch_size=batch_size,
            num_workers=num_workers,
            custom_transforms=custom_transforms,
            custom_target_transforms=custom_target_transforms,
        )

    def prepare_data(self) -> None:
        r"""Download the original CIFAR-100 dataset if haven't."""
        # just download
        CIFAR100(root=self.root, train=True, download=True)
        CIFAR100(root=self.root, train=False, download=True)

        pylogger.debug(
            "The original CIFAR-100 dataset has been downloaded to %s.", self.root
        )

    def train_and_val_dataset(self) -> tuple[Dataset, Dataset]:
        r"""Get the training and validation dataset of task `self.task_id`.

        **Returns:**
        - **train_and_val_dataset** (`tuple[Dataset, Dataset]`): the train and validation dataset of task `self.task_id`.
        """
        dataset_train_and_val = self.get_class_subset(
            CIFAR100(
                root=self.root,
                train=True,
                transform=self.train_and_val_transforms(to_tensor=True),
                download=False,
            )
        )
        dataset_train_and_val.target_transform = self.target_transforms()

        return random_split(
            dataset_train_and_val,
            lengths=[1 - self.validation_percentage, self.validation_percentage],
        )

    def test_dataset(self) -> Dataset:
        r"""Get the test dataset of task `self.task_id`.

        **Returns:**
        - **test_dataset** (`Dataset`): the test dataset of task `self.task_id`.
        """
        dataset_test = self.get_class_subset(
            CIFAR100(
                root=self.root,
                train=False,
                transform=self.test_transforms(to_tensor=True),
                download=False,
            )
        )
        dataset_test.target_transform = self.target_transforms()

        return dataset_test



"""The submodule in `utils` for transforming tensors."""

__all__ = ["min_max_normalise"]


from torch import Tensor


def min_max_normalise(
    tensor: Tensor, dim: int | None = None, epsilon: float = 1e-8
) -> Tensor:
    r"""Normalise the tensor using min-max normalisation.

    **Args:**
    - **tensor** (`Tensor`): the input tensor to normalise.
    - **dim** (`int` | `None`): the dimension to normalise along. If `None`, normalise the whole tensor.
    - **epsilon** (`float`): the epsilon value to avoid division by zero.

    **Returns:**
    - **tensor** (`Tensor`): the normalised tensor.
    """
    min_val = (
        tensor.min(dim=dim, keepdim=True).values if dim is not None else tensor.min()
    )
    max_val = (
        tensor.max(dim=dim, keepdim=True).values if dim is not None else tensor.max()
    )

    return (tensor - min_val) / (max_val - min_val + epsilon)



"""The submodule in `utils` for plotting utils."""

__all__ = [
    "plot_test_ave_acc_curve_from_csv",
    "plot_test_acc_matrix_from_csv",
    "plot_test_ave_loss_cls_curve_from_csv",
    "plot_test_loss_cls_matrix_from_csv",
    "plot_hat_mask",
]


import os

import pandas as pd
from matplotlib import pyplot as plt
from torch import Tensor


def plot_test_ave_acc_curve_from_csv(
    csv_path: str, task_id: int, plot_path: str
) -> None:
    """Plot the test average accuracy curve over different training tasks from saved csv file and save the plot to the designated directory.

    **Args:**
    - **csv_path** (`str`): the path to the csv file where the `utils.update_test_acc_to_csv()` saved the test accuracy metric.
    - **task_id** (`int`): plot the test average accuracy metric from task 1 to `task_id`.
    - **plot_path** (`str`): the path to save plot. Better same as the output directory of the experiment. E.g. './outputs/expr_name/1970-01-01_00-00-00/ave_acc.png'.
    """
    data = pd.read_csv(csv_path)

    # plot the average accuracy curve over different training tasks
    fig, ax = plt.subplots(figsize=(16, 9))
    ax.plot(
        data["after_training_task"],
        data["average_accuracy"],
        marker="o",
        linewidth=2,
    )
    ax.set_xlabel("After training task $t$", fontsize=16)
    ax.set_xlabel("Average Accuracy (AA)", fontsize=16)
    ax.grid(True)
    xticks = [int(i) for i in range(1, task_id + 1)]
    yticks = [i * 0.05 for i in range(21)]
    ax.set_xticks(xticks)
    ax.set_yticks(yticks)
    ax.set_xticklabels(xticks, fontsize=16)
    ax.set_yticklabels([f"{tick:.2f}" for tick in yticks], fontsize=16)
    fig.savefig(plot_path)
    plt.close(fig)


def plot_test_acc_matrix_from_csv(csv_path: str, task_id: int, plot_path: str) -> None:
    """Plot the test accuracy matrix from saved csv file and save the plot to the designated directory.

    **Args:**
    - **csv_path** (`str`): the path to the csv file where the `utils.update_test_acc_to_csv()` saved the test accuracy metric.
    - **task_id** (`int`): plot the test accuracy metric from task 1 to `task_id`.
    - **plot_path** (`str`): the path to save plot. Better same as the output directory of the experiment. E.g. './outputs/expr_name/1970-01-01_00-00-00/acc_matrix.png'.
    """
    data = pd.read_csv(csv_path)

    # plot the accuracy matrix
    fig, ax = plt.subplots(
        figsize=(2 * (task_id + 1), 2 * (task_id + 1))
    )  # adaptive figure size
    cax = ax.imshow(
        data.drop(["after_training_task", "average_accuracy"], axis=1),
        interpolation="nearest",
        cmap="Greens",
    )
    colorbar = fig.colorbar(cax)
    yticks = colorbar.ax.get_yticks()
    colorbar.ax.set_yticks(yticks)
    colorbar.ax.set_yticklabels(
        [f"{tick:.2f}" for tick in yticks], fontsize=10 + task_id
    )  # adaptive font size

    for i in range(task_id + 1):
        for j in range(1, i + 1):
            ax.text(
                j - 1,
                i - 1,
                f'{data.loc[i - 1,f"test_on_task_{j}"]:.3f}',
                ha="center",
                va="center",
                color="black",
                fontsize=10 + task_id,  # adaptive font size
            )

    ax.set_xticks(range(task_id))
    ax.set_yticks(range(task_id))
    ax.set_xticklabels(
        range(1, task_id + 1), fontsize=10 + task_id
    )  # adaptive font size
    ax.set_yticklabels(
        range(1, task_id + 1), fontsize=10 + task_id
    )  # adaptive font size

    # Labeling the axes
    ax.set_xlabel("Testing on task τ", fontsize=10 + task_id)  # adaptive font size
    ax.set_ylabel("After training task t", fontsize=10 + task_id)  # adaptive font size
    fig.savefig(plot_path)
    plt.close(fig)


def plot_test_ave_loss_cls_curve_from_csv(
    csv_path: str, task_id: int, plot_path: str
) -> None:
    """Plot the test average classification loss curve over different training tasks from saved csv file and save the plot to the designated directory.

    **Args:**
    - **csv_path** (`str`): the path to the csv file where the `utils.update_loss_cls_to_csv()` saved the test classification loss metric.
    - **task_id** (`int`): plot the test average accuracy metric from task 1 to `task_id`.
    - **plot_path** (`str`): the path to save plot. Better same as the output directory of the experiment. E.g. './outputs/expr_name/1970-01-01_00-00-00/ave_loss_cls.png'.
    """
    data = pd.read_csv(csv_path)

    # plot the average accuracy curve over different training tasks
    fig, ax = plt.subplots(figsize=(16, 9))
    ax.plot(
        data["after_training_task"],
        data["average_classification_loss"],
        marker="o",
        linewidth=2,
    )
    ax.set_xlabel("After training task $t$", fontsize=16)
    ax.set_xlabel("Average Classification Loss", fontsize=16)
    ax.grid(True)

    xticks = [int(i) for i in range(1, task_id + 1)]
    yticks = [
        i * 0.5 for i in range(int(data["average_classification_loss"].max() / 0.5) + 1)
    ]
    ax.set_xticks(xticks)
    ax.set_yticks(yticks)
    ax.set_xticklabels(xticks, fontsize=16)
    ax.set_yticklabels([f"{tick:.1f}" for tick in yticks], fontsize=16)
    fig.savefig(plot_path)
    plt.close(fig)


def plot_test_loss_cls_matrix_from_csv(
    csv_path: str, task_id: int, plot_path: str
) -> None:
    """Plot the test classification loss matrix from saved csv file and save the plot to the designated directory.

    **Args:**
    - **csv_path** (`str`): the path to the csv file where the `utils.update_loss_cls_to_csv()` saved the test classification loss metric.
    - **task_id** (`int`): plot the test classification loss metric from task 1 to `task_id`.
    - **plot_path** (`str`): the path to save plot. Better same as the output directory of the experiment. E.g. './outputs/expr_name/1970-01-01_00-00-00/loss_cls_matrix.png'.
    """
    data = pd.read_csv(csv_path)

    # plot the accuracy matrix
    fig, ax = plt.subplots(
        figsize=(2 * (task_id + 1), 2 * (task_id + 1))
    )  # adaptive figure size
    cax = ax.imshow(
        data.drop(["after_training_task", "average_classification_loss"], axis=1),
        interpolation="nearest",
        cmap="Greens",
    )
    colorbar = fig.colorbar(cax)
    yticks = colorbar.ax.get_yticks()
    colorbar.ax.set_yticks(yticks)
    colorbar.ax.set_yticklabels(
        [f"{tick:.2f}" for tick in yticks], fontsize=10 + task_id
    )  # adaptive font size

    for i in range(task_id + 1):
        for j in range(1, i + 1):
            ax.text(
                j - 1,
                i - 1,
                f'{data.loc[i - 1,f"test_on_task_{j}"]:.3f}',
                ha="center",
                va="center",
                color="black",
                fontsize=10 + task_id,  # adaptive font size
            )
    ax.set_xticks(range(task_id))
    ax.set_yticks(range(task_id))

    ax.set_xticklabels(
        range(1, task_id + 1), fontsize=10 + task_id
    )  # adaptive font size
    ax.set_yticklabels(
        range(1, task_id + 1), fontsize=10 + task_id
    )  # adaptive font size

    # Labeling the axes
    ax.set_xlabel("Testing on task τ", fontsize=10 + task_id)  # adaptive font size
    ax.set_ylabel("After training task t", fontsize=10 + task_id)  # adaptive font size
    fig.savefig(plot_path)
    plt.close(fig)


def plot_hat_mask(
    mask: dict[str, Tensor], plot_dir: str, task_id: int, step: int | None = None
) -> None:
    """Plot mask in [HAT (Hard Attention to the Task)](http://proceedings.mlr.press/v80/serra18a)) algorithm. This includes the mask and cumulative mask.

    **Args:**
    - **mask** (`dict[str, Tensor]`): the hard attention (whose values are 0 or 1) mask. Key (`str`) is layer name, value (`Tensor`) is the mask tensor. The mask tensor has size (number of units).
    - **plot_dir** (`str`): the directory to save plot. Better same as the output directory of the experiment.
    - **task_id** (`int`): the task ID of the mask to be plotted. This is to form the plot name.
    - **step** (`int`): the training step (batch index) of the mask to be plotted. Apply to the training mask only. This is to form the plot name. Keep `None` for not showing the step in the plot name.
    """

    for layer_name, m in mask.items():
        layer_name = layer_name.replace(
            "/", "."
        )  # the layer name contains '/', which is not allowed in the file name. We replace it back with '.'.

        m = m.view(1, -1)  # reshape the 1D mask to 2D so can be plotted by image show

        fig = plt.figure()
        plt.imshow(
            m.detach().cpu(), aspect=10, cmap="Greys"
        )  # can only convert to tensors in CPU to numpy arrays
        plt.yticks()  # hide yticks
        plt.colorbar()
        if step:
            plot_name = f"{layer_name}_task{task_id}_step{step}.png"
        else:
            plot_name = f"{layer_name}_task{task_id}.png"
        plot_path = os.path.join(plot_dir, plot_name)
        fig.savefig(plot_path)
        plt.close(fig)



r"""

# Continual Learning Datasets

This submodule provides the **continual learning datasets** that can be used in CLArena. 

Please note that this is an API documantation. Please refer to the main documentation pages for more information about the CL datasets and how to configure and implement them:

- [**Configure CL Dataset**](https://pengxiang-wang.com/projects/continual-learning-arena/docs/configure-your-experiment/cl-dataset)
- [**Implement Your CL Dataset Class**](https://pengxiang-wang.com/projects/continual-learning-arena/docs/implement-your-cl-modules/cl-dataset)
- [**A Beginners' Guide to Continual Learning (CL Dataset)**](https://pengxiang-wang.com/posts/continual-learning-beginners-guide#sec-CL-dataset)

The datasets are implemented as subclasses of `CLDataset` classes, which are the base class for all continual learning datasets in CLArena.

- `CLDataset`: The base class for continual learning datasets.
- `CLPermutedDataset`: The base class for permuted continual learning datasets. A child class of `CLDataset`.

"""

from .base import CLClassMapping, CLDataset, CLPermutedDataset, CLSplitDataset, Permute
from .permuted_mnist import PermutedMNIST
from .split_cifar100 import SplitCIFAR100

__all__ = [
    "CLDataset",
    "CLPermutedDataset",
    "CLSplitDataset",
    "CLClassMapping",
    "Permute",
    "permuted_mnist",
    "split_cifar100",
]



r"""
The submodule in `cl_algorithms` for [LwF (Learning without Forgetting) algorithm](https://ieeexplore.ieee.org/abstract/document/8107520).
"""

__all__ = ["LwF"]

import logging
from copy import deepcopy
from typing import Any

import torch
from torch import Tensor, nn

from clarena.backbones import CLBackbone
from clarena.cl_algorithms import Finetuning
from clarena.cl_algorithms.regularisers import DistillationReg
from clarena.cl_heads import HeadsCIL, HeadsTIL

# always get logger for built-in logging in each module
pylogger = logging.getLogger(__name__)


class LwF(Finetuning):
    r"""LwF (Learning without Forgetting) algorithm.

    [LwF (Learning without Forgetting, 2017)](https://ieeexplore.ieee.org/abstract/document/8107520) is a regularisation-based continual learning approach that constrains the feature output of the model to be similar to that of the previous tasks. From the perspective of knowledge distillation, it distills previous tasks models into the training process for new task in the regularisation term. It is a simple yet effective method for continual learning.

    We implement LwF as a subclass of Finetuning algorithm, as LwF has the same `forward()`, `validation_step()` and `test_step()` method as `Finetuning` class.
    """

    def __init__(
        self,
        backbone: CLBackbone,
        heads: HeadsTIL | HeadsCIL,
        distillation_reg_factor: float,
        distillation_reg_temparture: float,
    ) -> None:
        r"""Initialise the LwF algorithm with the network.

        **Args:**
        - **backbone** (`CLBackbone`): backbone network.
        - **heads** (`HeadsTIL` | `HeadsCIL`): output heads.
        - **distillation_reg_factor** (`float`): hyperparameter, the distillation regularisation factor. It controls the strength of preventing forgetting.
        - **distillation_reg_temparture** (`float`): hyperparameter, the temperature in the distillation regularisation. It controls the softness of the labels that the student model (here is the current model) learns from the teacher models (here are the previous models), thereby controlling the strength of the distillation. It controls the strength of preventing forgetting.
        """
        Finetuning.__init__(self, backbone=backbone, heads=heads)

        self.previous_task_backbones: dict[str, nn.Module] = {}
        r"""Store the backbone models of the previous tasks. Keys are task IDs (string type) and values are the corresponding models. Each model is a `nn.Module` backbone after the corresponding previous task was trained.
        
        Some would argue that since we could store the model of the previous tasks, why don't we test the task directly with the stored model, instead of doing the less easier LwF thing? The thing is, LwF only uses the model of the previous tasks to train current and future tasks, which aggregate them into a single model. Once the training of the task is done, the storage for those parameters can be released. However, this make the future tasks not able to use LwF anymore, which is a disadvantage for LwF.
        """

        self.distillation_reg_factor = distillation_reg_factor
        r"""Store distillation regularisation factor."""
        self.distillation_reg_temperature = distillation_reg_temparture
        r"""Store distillation regularisation temperature."""
        self.distillation_reg = DistillationReg(
            factor=distillation_reg_factor,
            temperature=distillation_reg_temparture,
            distance="cross_entropy",
        )
        r"""Initialise and store the distillation regulariser."""

        LwF.sanity_check(self)

    def sanity_check(self) -> None:
        r"""Check the sanity of the arguments.

        **Raises:**
        - **ValueError**: If the regularisation factor and distillation temperature is not positive.
        """

        if self.distillation_reg_factor <= 0:
            raise ValueError(
                f"The distillation regularisation factor should be positive, but got {self.distillation_reg_factor}."
            )

        if self.distillation_reg_temperature <= 0:
            raise ValueError(
                f"The distillation regularisation temperature should be positive, but got {self.distillation_reg_temperature}."
            )

    def training_step(self, batch: Any) -> dict[str, Tensor]:
        r"""Training step for current task `self.task_id`.

        **Args:**
        - **batch** (`Any`): a batch of training data.

        **Returns:**
        - **outputs** (`dict[str, Tensor]`): a dictionary contains loss and other metrics from this training step. Key (`str`) is the metrics name, value (`Tensor`) is the metrics. Must include the key 'loss' which is total loss in the case of automatic optimization, according to PyTorch Lightning docs.
        """
        x, y = batch

        # classification loss. See equation (1) in the [LwF paper](https://ieeexplore.ieee.org/abstract/document/8107520).
        logits, hidden_features = self.forward(x, stage="train", task_id=self.task_id)
        loss_cls = self.criterion(logits, y)

        # regularisation loss. See equation (2) (3) in the [LwF paper](https://ieeexplore.ieee.org/abstract/document/8107520).
        loss_reg = 0.0
        for previous_task_id in range(1, self.task_id):
            # sum over all previous models, because [LwF paper](https://ieeexplore.ieee.org/abstract/document/8107520) says: "If there are multiple old tasks, or if an old task is multi-label classification, we take the sum of the loss for each old task and label."

            # get the teacher logits for this batch, which is from the current model (to previous output head)
            student_feature, _ = self.backbone(
                x, stage="train", task_id=previous_task_id
            )
            with torch.no_grad():  # stop updating the previous heads
                student_logits = self.heads(student_feature, task_id=previous_task_id)

            # get the teacher logits for this batch, which is from the previous model
            previous_backbone = self.previous_task_backbones[previous_task_id]
            with torch.no_grad():  # stop updating the previous backbones and heads
                teacher_feature, _ = previous_backbone(
                    x, stage="test", task_id=previous_task_id
                )

                teacher_logits = self.heads(teacher_feature, task_id=previous_task_id)

            loss_reg += self.distillation_reg(
                student_logits=student_logits,
                teacher_logits=teacher_logits,
            )

        if self.task_id != 1:
            loss_reg /= (
                self.task_id
            )  # average over tasks to avoid linear increase of the regularisation loss

        # total loss
        loss = loss_cls + loss_reg

        # accuracy of the batch
        acc = (logits.argmax(dim=1) == y).float().mean()

        return {
            "loss": loss,  # Return loss is essential for training step, or backpropagation will fail
            "loss_cls": loss_cls,
            "loss_reg": loss_reg,
            "acc": acc,
            "hidden_features": hidden_features,
        }

    def on_train_end(self) -> None:
        r"""Store the backbone model after the training of a task.

        The model is stored in `self.previous_task_backbones` for constructing the regularisation loss in the future tasks.
        """
        previous_backbone = deepcopy(self.backbone)
        previous_backbone.eval()  # set the store model to evaluation mode to prevent updating
        self.heads.heads[
            f"{self.task_id}"
        ].eval()  # set the store model to evaluation mode to prevent updating
        self.previous_task_backbones[self.task_id] = previous_backbone



r"""The submodule in `cl_heads` for TIL heads."""

__all__ = ["HeadsTIL"]

from torch import Tensor, nn


class HeadsTIL(nn.Module):
    r"""The output heads for Task-Incremental Learning (TIL). Independent head assigned to each TIL task takes the output from backbone network and forwards it into logits for predicting classes of the task."""

    def __init__(self, input_dim: int) -> None:
        r"""Initializes TIL heads object with no heads.

        **Args:**
        - **input_dim** (`int`): the input dimension of the heads. Must be equal to the `output_dim` of the connected backbone.
        """
        nn.Module.__init__(self)

        self.heads: nn.ModuleDict = nn.ModuleDict()  # initially no heads
        r"""TIL output heads are stored independently in a `ModuleDict`. Keys are task IDs (string type) and values are the corresponding `nn.Linear` heads. We use `ModuleDict` rather than `dict` to make sure `LightningModule` can track these model parameters for the purpose of, such as automatically to device, recorded in model summaries.
        
        Note that the task IDs must be string type in order to let `LightningModule` identify this part of the model. """
        self.input_dim: int = input_dim
        r"""Store the input dimension of the heads. Used when creating new heads."""

        self.task_id: int
        r"""Task ID counter indicating which task is being processed. Self updated during the task loop."""

    def setup_task_id(self, task_id: int, num_classes_t: int) -> None:
        r"""Create the output head when task `task_id` arrives if there's no. This must be done before `forward()` is called.

        **Args:**
        - **task_id** (`int`): the target task ID.
        - **num_classes_t** (`int`): the number of classes in the task.
        """
        self.task_id = task_id
        if self.task_id not in self.heads.keys():
            self.head_t = nn.Linear(self.input_dim, num_classes_t)
            self.heads[f"{self.task_id}"] = self.head_t

    def forward(self, feature: Tensor, task_id: int) -> Tensor:
        r"""The forward pass for data from task `task_id`. A head is selected according to the task_id and the feature is passed through the head.

        **Args:**
        - **feature** (`Tensor`): the feature tensor from the backbone network.
        - **task_id** (`int`): the task ID where the data are from, which is provided by task-incremental setting.

        **Returns:**
        - **logits** (`Tensor`): the output logits tensor.
        """

        head_t = self.heads[f"{task_id}"]
        logits = head_t(feature)

        return logits



r"""
The submodule in `cl_algorithms` for CL algorithm bases.
"""

__all__ = ["CLAlgorithm"]

import logging

from lightning import LightningModule
from torch import nn
from torch.optim import Optimizer

from clarena.backbones import CLBackbone
from clarena.cl_heads import HeadsCIL, HeadsTIL

# always get logger for built-in logging in each module
pylogger = logging.getLogger(__name__)


class CLAlgorithm(LightningModule):
    r"""The base class of continual learning algorithms, inherited from `LightningModule`."""

    def __init__(
        self,
        backbone: CLBackbone,
        heads: HeadsTIL | HeadsCIL,
    ) -> None:
        r"""Initialise the CL algorithm with the network.

        **Args:**
        - **backbone** (`CLBackbone`): backbone network.
        - **heads** (`HeadsTIL` | `HeadsCIL`): output heads.
        """
        LightningModule.__init__(self)

        self.backbone: CLBackbone = backbone
        r"""Store the backbone network."""
        self.heads: HeadsTIL | HeadsCIL = heads
        r"""Store the output heads."""
        self.optimizer: Optimizer
        r"""Store the optimizer object (partially initialised) for the backpropagation of task `self.task_id`. Will be equipped with parameters in `configure_optimizers()`."""
        self.criterion = nn.CrossEntropyLoss()
        r"""The loss function bewteen the output logits and the target labels. Default is cross-entropy loss."""

        self.task_id: int
        r"""Task ID counter indicating which task is being processed. Self updated during the task loop."""

        CLAlgorithm.sanity_check(self)

    def sanity_check(self) -> None:
        r"""Check the sanity of the arguments.

        **Raises:**
        - **ValueError**: if the `output_dim` of backbone network is not equal to the `input_dim` of CL heads.
        """
        if self.backbone.output_dim != self.heads.input_dim:
            raise ValueError(
                "The output_dim of backbone network should be equal to the input_dim of CL heads!"
            )

    def setup_task_id(
        self, task_id: int, num_classes_t: int, optimizer: Optimizer
    ) -> None:
        r"""Set up which task's dataset the CL experiment is on. This must be done before `forward()` method is called.

        **Args:**
        - **task_id** (`int`): the target task ID.
        - **num_classes_t** (`int`): the number of classes in the task.
        - **optimizer** (`Optimizer`): the optimizer object (partially initialised) for the task `self.task_id`.
        """
        self.task_id = task_id
        self.heads.setup_task_id(task_id, num_classes_t)
        self.optimizer = optimizer

    def configure_optimizers(self) -> Optimizer:
        r"""
        Configure optimizer hooks by Lightning.
        See [Lightning docs](https://lightning.ai/docs/pytorch/stable/common/lightning_module.html#configure-optimizers) for more details.
        """
        # finish partially initialised optimizer by specifying model parameters. The `parameters()` method of this `CLAlrogithm` (inherited from `LightningModule`) returns both backbone and heads parameters
        return self.optimizer(params=self.parameters())



r"""
The submodule in `cl_datasets` for CL dataset bases.
"""

__all__ = ["CLDataset", "CLPermutedDataset", "CLClassMapping", "Permute"]

import logging
from abc import abstractmethod
from typing import Any, Callable

import torch
from lightning import LightningDataModule
from torch.utils.data import DataLoader, Dataset
from torchvision import transforms

# always get logger for built-in logging in each module
pylogger = logging.getLogger(__name__)


class CLDataset(LightningDataModule):
    r"""The base class of continual learning datasets, inherited from `LightningDataModule`."""

    def __init__(
        self,
        root: str,
        num_tasks: int,
        validation_percentage: float,
        batch_size: int = 1,
        num_workers: int = 10,
        custom_transforms: Callable | transforms.Compose | None = None,
        custom_target_transforms: Callable | transforms.Compose | None = None,
    ) -> None:
        r"""Initialise the CL dataset object providing the root where data files live.

        **Args:**
        - **root** (`str`): the root directory where the original data files for constructing the CL dataset physically live.
        - **num_tasks** (`int`): the maximum number of tasks supported by the CL dataset.
        - **validation_percentage** (`float`): the percentage to randomly split some of the training data into validation data.
        - **batch_size** (`int`): The batch size in train, val, test dataloader.
        - **num_workers** (`int`): the number of workers for dataloaders.
        - **custom_transforms** (`transform` or `transforms.Compose` or `None`): the custom transforms to apply to ONLY TRAIN dataset. Can be a single transform, composed transforms or no transform. `ToTensor()`, normalise, permute and so on are not included.
        - **custom_target_transforms** (`transform` or `transforms.Compose` or `None`): the custom target transforms to apply to dataset labels. Can be a single transform, composed transforms or no transform. CL class mapping is not included.
        """
        LightningDataModule.__init__(self)

        self.root: str = root
        r"""Store the root directory of the original data files. Used when constructing the dataset."""
        self.num_tasks: int = num_tasks
        r"""Store the maximum number of tasks supported by the dataset."""
        self.validation_percentage: float = validation_percentage
        r"""Store the percentage to randomly split some of the training data into validation data."""
        self.batch_size: int = batch_size
        r"""Store the batch size. Used when constructing train, val, test dataloader."""
        self.num_workers: int = num_workers
        r"""Store the number of workers. Used when constructing train, val, test dataloader."""
        self.custom_transforms: Callable | transforms.Compose | None = custom_transforms
        r"""Store the custom transforms other than the basics. Used when constructing the dataset."""
        self.custom_target_transforms: Callable | transforms.Compose | None = (
            custom_target_transforms
        )
        r"""Store the custom target transforms other than the CL class mapping. Used when constructing the dataset."""

        self.task_id: int
        r"""Task ID counter indicating which task is being processed. Self updated during the task loop."""
        self.cl_paradigm: str
        r"""Store the continual learning paradigm, either 'TIL' (Task-Incremental Learning) or 'CIL' (Class-Incremental Learning). Gotten from `set_cl_paradigm` and used to define the CL class map."""

        self.cl_class_map_t: dict[str | int, int]
        r"""Store the CL class map for the current task `self.task_id`. """
        self.cl_class_mapping_t: Callable
        r"""Store the CL class mapping transform for the current task `self.task_id`. """

        self.dataset_train: object
        r"""The training dataset object. Can be a PyTorch Dataset object or any other dataset object."""
        self.dataset_val: object
        r"""The validation dataset object. Can be a PyTorch Dataset object or any other dataset object."""
        self.dataset_test: dict[str, object] = {}
        r"""The dictionary to store test dataset object. Keys are task IDs (string type) and values are the dataset objects. Can be PyTorch Dataset objects or any other dataset objects."""

        CLDataset.sanity_check(self)

    def sanity_check(self) -> None:
        r"""Check the sanity of the arguments.

        **Raises:**
        - **ValueError**: when the `validation_percentage` is not in the range of 0-1.
        """
        if not 0.0 < self.validation_percentage < 1.0:
            raise ValueError("The validation_percentage should be 0-1.")

    @abstractmethod
    def cl_class_map(self, task_id: int) -> dict[str | int, int]:
        r"""The mapping of classes of task `task_id` to fit continual learning settings `self.cl_paradigm`. It must be implemented by subclasses.

        **Args:**
        - **task_id** (`int`): The task ID to query CL class map.

        **Returns:**
        - **cl_class_map**(`dict[str | int, int]`): the CL class map of the task. Key is original class label, value is integer class label for continual learning.
            - If `self.cl_paradigm` is 'TIL', the mapped class labels of a task should be continuous integers from 0 to the number of classes.
            - If `self.cl_paradigm` is 'CIL', the mapped class labels of a task should be continuous integers from the number of classes of previous tasks to the number of classes of the current task.
        """

    @abstractmethod
    def prepare_data(self) -> None:
        r"""Use this to download and prepare data. It must be implemented by subclasses, regulated by `LightningDatamodule`."""

    def setup(self, stage: str) -> None:
        r"""Set up the dataset for different stages.

        **Args:**
        - **stage** (`str`): the stage of the experiment. Should be one of the following:
            - 'fit' or 'validation': training and validation dataset of current task `self.task_id` should be assigned to `self.dataset_train` and `self.dataset_val`.
            - 'test': a list of test dataset of all seen tasks (from task 0 to `self.task_id`) should be assigned to `self.dataset_test`.
        """
        if stage == "fit" or "validate":

            pylogger.debug(
                "Construct train and validation dataset for task %d...", self.task_id
            )
            self.dataset_train, self.dataset_val = self.train_and_val_dataset()
            self.dataset_train.target_transform = (
                self.target_transforms()
            )  # apply target transform after potential class split
            self.dataset_val.target_transform = (
                self.target_transforms()
            )  # apply target transform after potential class split
            pylogger.debug(
                "Train and validation dataset for task %d are ready.", self.task_id
            )

        if stage == "test":

            pylogger.debug("Construct test dataset for task %d...", self.task_id)
            self.dataset_test[f"{self.task_id}"] = self.test_dataset()
            self.dataset_test[f"{self.task_id}"].target_transform = (
                self.target_transforms()
            )  # apply target transform after potential class split
            pylogger.debug("Test dataset for task %d are ready.", self.task_id)

    def setup_task_id(self, task_id: int) -> None:
        r"""Set up which task's dataset the CL experiment is on. This must be done before `setup()` method is called.

        **Args:**
        - **task_id** (`int`): the target task ID.
        """
        self.task_id = task_id

        self.cl_class_map_t = self.cl_class_map(task_id)
        self.cl_class_mapping_t = CLClassMapping(self.cl_class_map_t)

    def set_cl_paradigm(self, cl_paradigm: str) -> None:
        r"""Set the continual learning paradigm to `self.cl_paradigm`. It is used to define the CL class map.

        **Args:**
        - **cl_paradigm** (`str`): the continual learning paradigmeither 'TIL' (Task-Incremental Learning) or 'CIL' (Class-Incremental Learning).
        """
        self.cl_paradigm = cl_paradigm

    @abstractmethod
    def mean(self, task_id: int) -> tuple[float]:
        r"""The mean values for normalisation of task `task_id`. Used when constructing the dataset.

        **Returns:**
        - **mean** (`tuple[float]`): the mean values for normalisation.
        """

    @abstractmethod
    def std(self, task_id: int) -> tuple[float]:
        r"""The standard deviation values for normalisation of task `task_id`. Used when constructing the dataset.

        **Returns:**
        - **std** (`tuple[float]`): the standard deviation values for normalisation.
        """

    def train_and_val_transforms(self, to_tensor: bool) -> transforms.Compose:
        r"""Transforms generator for train and validation dataset incorporating the custom transforms with basic transforms like `normalisation` and `ToTensor()`. It is a handy tool to use in subclasses when constructing the dataset.

        **Args:**
        - **to_tensor** (`bool`): whether to include `ToTensor()` transform.

        **Returns:**
        - **train_and_val_transforms** (`transforms.Compose`): the composed training transforms.
        """

        return transforms.Compose(
            list(
                filter(
                    None,
                    [
                        transforms.ToTensor() if to_tensor else None,
                        self.custom_transforms,
                        transforms.Normalize(
                            self.mean(self.task_id), self.std(self.task_id)
                        ),
                    ],
                )
            )
        )  # the order of transforms matters

    def test_transforms(self, to_tensor: bool) -> transforms.Compose:
        r"""Transforms generator for test dataset. Only basic transforms like `normalisation` and `ToTensor()` are included. It is a handy tool to use in subclasses when constructing the dataset.

        **Args:**
        - **to_tensor** (`bool`): whether to include `ToTensor()` transform.

        **Returns:**
        - **test_transforms** (`transforms.Compose`): the composed training transforms.
        """

        return transforms.Compose(
            list(
                filter(
                    None,
                    [
                        transforms.ToTensor() if to_tensor else None,
                        transforms.Normalize(
                            self.mean(self.task_id), self.std(self.task_id)
                        ),
                    ],
                )
            )
        )  # the order of transforms matters

    def target_transforms(self) -> transforms.Compose:
        r"""The target transform for the dataset. It is a handy tool to use in subclasses when constructing the dataset.

        **Args:**
        - **target** (`Tensor`): the target tensor.

        **Returns:**
        - **target_transforms** (`transforms.Compose`): the transformed target tensor.
        """

        return transforms.Compose(
            list(
                filter(
                    None,
                    [
                        self.custom_target_transforms,
                        self.cl_class_mapping_t,
                    ],
                )
            )
        )  # the order of transforms matters

    @abstractmethod
    def train_and_val_dataset(self) -> Any:
        r"""Get the training and validation dataset of task `self.task_id`. It must be implemented by subclasses.

        **Returns:**
        - **train_and_val_dataset** (`Any`): the train and validation dataset of task `self.task_id`.
        """

    @abstractmethod
    def test_dataset(self) -> Any:
        """Get the test dataset of task `self.task_id`. It must be implemented by subclasses.

        **Returns:**
        - **train_and_val_dataset** (`Any`): the test dataset of task `self.task_id`.
        """

    def train_dataloader(self) -> DataLoader:
        r"""DataLoader generator for stage train of task `self.task_id`. It is automatically called before training.

        **Returns:**
        - **train_dataloader** (`Dataloader`): the train DataLoader of task `self.task_id`.
        """

        pylogger.debug("Construct train dataloader for task %d...", self.task_id)

        return DataLoader(
            dataset=self.dataset_train,
            batch_size=self.batch_size,
            shuffle=True,  # shuffle train batch to prevent overfitting
            num_workers=self.num_workers,
            persistent_workers=True,
        )

    def val_dataloader(self) -> DataLoader:
        r"""DataLoader generator for stage validate. It is automatically called before validation.

        **Returns:**
        - **val_dataloader** (`Dataloader`): the validation DataLoader of task `self.task_id`.
        """

        pylogger.debug("Construct validation dataloader for task %d...", self.task_id)

        return DataLoader(
            dataset=self.dataset_val,
            batch_size=self.batch_size,
            shuffle=False,  # don't have to shuffle val or test batch
            num_workers=self.num_workers,
            persistent_workers=True,
        )

    def test_dataloader(self) -> dict[int, DataLoader]:
        r"""DataLoader generator for stage test. It is automatically called before testing.

        **Returns:**
        - **test_dataloader** (`dict[int, DataLoader]`): the test DataLoader dict of `self.task_id` and all tasks before (as the test is conducted on all seen tasks). Keys are task IDs (integer type) and values are the DataLoaders.
        """

        pylogger.debug("Construct test dataloader for task %d...", self.task_id)

        return {
            task_id: DataLoader(
                dataset=dataset_test,
                batch_size=self.batch_size,
                shuffle=False,  # don't have to shuffle val or test batch
                num_workers=self.num_workers,
                persistent_workers=True,  # speed up the dataloader worker initialization
            )
            for task_id, dataset_test in self.dataset_test.items()
        }


class CLPermutedDataset(CLDataset):
    r"""The base class of continual learning datasets which are constructed as permutations from an original dataset, inherited from `CLDataset`."""

    num_classes: int
    r"""The number of classes in the original dataset before permutation. It must be provided in subclasses."""

    img_size: torch.Size
    r"""The size of images in the original dataset before permutation. Used when constructing permutation operations. It must be provided in subclasses."""

    mean_original: tuple[float]
    r"""The mean values for normalisation. It must be provided in subclasses."""

    std_original: tuple[float]
    r"""The standard deviation values for normalisation. It must be provided in subclasses."""

    def __init__(
        self,
        root: str,
        num_tasks: int,
        validation_percentage: float,
        batch_size: int = 1,
        num_workers: int = 10,
        custom_transforms: Callable | transforms.Compose | None = None,
        custom_target_transforms: Callable | transforms.Compose | None = None,
        permutation_mode: str = "first_channel_only",
        permutation_seeds: list[int] | None = None,
    ) -> None:
        r"""Initialise the CL dataset object providing the root where data files live.

        **Args:**
        - **root** (`str`): the root directory where the original data files for constructing the CL dataset physically live.
        - **num_tasks** (`int`): the maximum number of tasks supported by the CL dataset.
        - **validation_percentage** (`float`): the percentage to randomly split some of the training data into validation data.
        - **batch_size** (`int`): The batch size in train, val, test dataloader.
        - **num_workers** (`int`): the number of workers for dataloaders.
        - **custom_transforms** (`transform` or `transforms.Compose` or `None`): the custom transforms to apply to ONLY TRAIN dataset. Can be a single transform, composed transforms or no transform. `ToTensor()`, normalise, permute and so on are not included.
        - **custom_target_transforms** (`transform` or `transforms.Compose` or `None`): the custom target transforms to apply to dataset labels. Can be a single transform, composed transforms or no transform. CL class mapping is not included.
        - **permutation_mode** (`str`): the mode of permutation, should be one of the following:
            1. 'all': permute all pixels.
            2. 'by_channel': permute channel by channel separately. All channels are applied the same permutation order.
            3. 'first_channel_only': permute only the first channel.
        - **permutation_seeds** (`list[int]` or `None`): the seeds for permutation operations used to construct tasks. Make sure it has the same number of seeds as `num_tasks`. Default is None, which creates a list of seeds from 1 to `num_tasks`.
        """
        CLDataset.__init__(
            self,
            root=root,
            num_tasks=num_tasks,
            validation_percentage=validation_percentage,
            batch_size=batch_size,
            num_workers=num_workers,
            custom_transforms=custom_transforms,
            custom_target_transforms=custom_target_transforms,
        )

        self.permutation_mode: str = permutation_mode
        r"""Store the mode of permutation. Used when permutation operations used to construct tasks. """

        self.permutation_seeds: list[int] = (
            permutation_seeds if permutation_seeds else list(range(num_tasks))
        )
        r"""Store the permutation seeds for all tasks. Use when permutation operations used to construct tasks. """

        self.permutation_seed_t: int
        r"""Store the permutation seed for the current task `self.task_id`."""
        self.permute_t: Permute
        r"""Store the permutation transform for the current task `self.task_id`. """

        CLPermutedDataset.sanity_check(self)

    def sanity_check(self) -> None:
        r"""Check the sanity of the arguments.

        **Raises:**
        - **ValueError**: when the `permutation_seeds` is not equal to `num_tasks`, or the `permutation_mode` is not one of the valid options.
        """
        if self.permutation_seeds and self.num_tasks != len(self.permutation_seeds):
            raise ValueError(
                "The number of permutation seeds is not equal to number of tasks!"
            )
        if self.permutation_mode not in ["all", "by_channel", "first_channel_only"]:
            raise ValueError(
                "The permutation_mode should be one of 'all', 'by_channel', 'first_channel_only'."
            )

    def cl_class_map(self, task_id: int) -> dict[str | int, int]:
        r"""The mapping of classes of task `task_id` to fit continual learning settings `self.cl_paradigm`.

        **Args:**
        - **task_id** (`int`): The task ID to query CL class map.

        **Returns:**
        - **cl_class_map**(`dict[str | int, int]`): the CL class map of the task. Key is original class label, value is integer class label for continual learning.
            - If `self.cl_paradigm` is 'TIL', the mapped class labels of a task should be continuous integers from 0 to the number of classes.
            - If `self.cl_paradigm` is 'CIL', the mapped class labels of a task should be continuous integers from the number of classes of previous tasks to the number of classes of the current task.
        """
        if self.cl_paradigm == "TIL":
            return {i: i for i in range(self.num_classes)}
        if self.cl_paradigm == "CIL":
            return {
                i: i + (task_id - 1) * self.num_classes for i in range(self.num_classes)
            }

    def setup_task_id(self, task_id: int) -> None:
        r"""Set up which task's dataset the CL experiment is on. This must be done before `setup()` method is called.

        **Args:**
        - **task_id** (`int`): the target task ID.
        """
        super().setup_task_id(task_id)

        self.permutation_seed_t = self.permutation_seeds[task_id - 1]
        self.permute_t = Permute(
            img_size=self.img_size,
            mode=self.permutation_mode,
            seed=self.permutation_seed_t,
        )

    def mean(self, task_id: int) -> tuple[float]:
        r"""The mean values for normalisation of task `task_id`. Used when constructing the dataset. In permuted CL dataset, the mean values are the same as the original dataset.

        **Returns:**
        - **mean** (`tuple[float]`): the mean values for normalisation.
        """
        return self.mean_original

    def std(self, task_id: int) -> tuple[float]:
        """The standard deviation values for normalisation of task `task_id`. Used when constructing the dataset. In permuted CL dataset, the mean values are the same as the original dataset.

        **Returns:**
        - **std** (`tuple[float]`): the standard deviation values for normalisation.
        """
        return self.std_original

    def train_and_val_transforms(self, to_tensor: bool) -> transforms.Compose:
        r"""Transforms generator for train and validation dataset incorporating the custom transforms with basic transforms like `normalisation` and `ToTensor()`. In permuted CL datasets, permute transform also applies. It is a handy tool to use in subclasses when constructing the dataset.

        **Args:**
        - **to_tensor** (`bool`): whether to include `ToTensor()` transform.

        **Returns:**
        - **train_and_val_transforms** (`transforms.Compose`): the composed training transforms.
        """

        return transforms.Compose(
            list(
                filter(
                    None,
                    [
                        transforms.ToTensor() if to_tensor else None,
                        self.permute_t,
                        self.custom_transforms,
                        transforms.Normalize(
                            self.mean(self.task_id), self.std(self.task_id)
                        ),
                    ],
                )
            )
        )  # the order of transforms matters

    def test_transforms(self, to_tensor: bool) -> transforms.Compose:
        r"""Transforms generator for test dataset. Only basic transforms like `normalisation` and `ToTensor()` are included. It is a handy tool to use in subclasses when constructing the dataset.

        **Args:**
        - **to_tensor** (`bool`): whether to include `ToTensor()` transform.

        **Returns:**
        - **test_transforms** (`transforms.Compose`): the composed training transforms.
        """

        return transforms.Compose(
            list(
                filter(
                    None,
                    [
                        transforms.ToTensor() if to_tensor else None,
                        self.permute_t,
                        transforms.Normalize(
                            self.mean(self.task_id), self.std(self.task_id)
                        ),
                    ],
                )
            )
        )  # the order of transforms matters


class CLSplitDataset(CLDataset):
    r"""The base class of continual learning datasets, which are constructed as permutations from an original dataset, inherited from `CLDataset`."""

    num_classes: int
    r"""The number of classes in the original dataset before permutation. It must be provided in subclasses."""

    mean_original: tuple[float]
    r"""The mean values for normalisation. It must be provided in subclasses."""

    std_original: tuple[float]
    r"""The standard deviation values for normalisation. It must be provided in subclasses."""

    def __init__(
        self,
        root: str,
        num_tasks: int,
        class_split: list[list[int]],
        validation_percentage: float,
        batch_size: int = 1,
        num_workers: int = 10,
        custom_transforms: Callable | transforms.Compose | None = None,
        custom_target_transforms: Callable | transforms.Compose | None = None,
    ) -> None:
        r"""Initialise the CL dataset object providing the root where data files live.

        **Args:**
        - **root** (`str`): the root directory where the original data files for constructing the CL dataset physically live.
        - **num_tasks** (`int`): the maximum number of tasks supported by the CL dataset.
        - **class_split** (`list[list[int]]`): the class split for each task. Each element in the list is a list of class labels (integers starting from 0) to split for a task.
        - **validation_percentage** (`float`): the percentage to randomly split some of the training data into validation data.
        - **batch_size** (`int`): The batch size in train, val, test dataloader.
        - **num_workers** (`int`): the number of workers for dataloaders.
        - **custom_transforms** (`transform` or `transforms.Compose` or `None`): the custom transforms to apply to ONLY TRAIN dataset. Can be a single transform, composed transforms or no transform. `ToTensor()`, normalise, permute and so on are not included.
        - **custom_target_transforms** (`transform` or `transforms.Compose` or `None`): the custom target transforms to apply to dataset labels. Can be a single transform, composed transforms or no transform. CL class mapping is not included.
        """
        CLDataset.__init__(
            self,
            root=root,
            num_tasks=num_tasks,
            validation_percentage=validation_percentage,
            batch_size=batch_size,
            num_workers=num_workers,
            custom_transforms=custom_transforms,
            custom_target_transforms=custom_target_transforms,
        )

        self.class_split = class_split
        r"""Store the class split for each task. Used when constructing the split dataset."""

        CLSplitDataset.sanity_check(self)

    def sanity_check(self) -> None:
        r"""Check the sanity of the arguments.

        **Raises:**
        - **ValueError**: when the length of `class_split` is not equal to `num_tasks`.
        - **ValueError**: when any of the lists in `class_split` has less than 2 elements. A classification task must have less than 2 classes.
        """
        if len(self.class_split) != self.num_tasks:
            raise ValueError(
                "The length of class split is not equal to number of tasks!"
            )
        if any(len(split) < 2 for split in self.class_split):
            raise ValueError("Each class split must contain at least 2 elements!")

    def cl_class_map(self, task_id: int) -> dict[str | int, int]:
        r"""The mapping of classes of task `task_id` to fit continual learning settings `self.cl_paradigm`.

        **Args:**
        - **task_id** (`int`): The task ID to query CL class map.

        **Returns:**
        - **cl_class_map**(`dict[str | int, int]`): the CL class map of the task. Key is original class label, value is integer class label for continual learning.
            - If `self.cl_paradigm` is 'TIL', the mapped class labels of a task should be continuous integers from 0 to the number of classes.
            - If `self.cl_paradigm` is 'CIL', the mapped class labels of a task should be continuous integers from the number of classes of previous tasks to the number of classes of the current task.
        """
        num_classes_t = len(self.class_split[task_id - 1])
        if self.cl_paradigm == "TIL":
            return {self.class_split[task_id - 1][i]: i for i in range(num_classes_t)}
        if self.cl_paradigm == "CIL":
            num_classes_previous = sum(
                [len(self.class_split[i]) for i in range(self.task_id - 1)]
            )
            return {
                self.class_split[task_id - 1][i]: i
                + (task_id - 1) * num_classes_previous
                for i in range(num_classes_t)
            }

    def mean(self, task_id: int) -> tuple[float]:
        r"""The mean values for normalisation of task `task_id`. Used when constructing the dataset. In split CL dataset, the mean values are the same as the original dataset.

        **Returns:**
        - **mean** (`tuple[float]`): the mean values for normalisation.
        """
        return self.mean_original

    def std(self, task_id: int) -> tuple[float]:
        r"""The standard deviation values for normalisation of task `task_id`. Used when constructing the dataset. In split CL dataset, the mean values are the same as the original dataset.

        **Returns:**
        - **std** (`tuple[float]`): he standard deviation values for normalisation.
        """
        return self.std_original

    def get_class_subset(self, dataset: Dataset) -> Dataset:
        r"""Provide a util method here to retrieve a subset from PyTorch Dataset of current classes of `self.task_id`. It could be useful when you constructing the split CL dataset.

        **Args:**
        - **dataset** (`Dataset`): the original dataset to retrieve subset from.

        **Returns:**
        - **subset** (`Dataset`): subset of original dataset in classes.
        """
        classes = self.class_split[self.task_id - 1]

        # get the indices of the dataset that belong to the classes
        idx = [i for i, (_, target) in enumerate(dataset) if target in classes]

        # subset the dataset by the indices, in-place operation
        dataset.data = dataset.data[idx]
        dataset.targets = [dataset.targets[i] for i in idx]

        return dataset


class CLClassMapping:
    r"""CL Class mapping to dataset labels. Used as a PyTorch target Transform."""

    def __init__(self, cl_class_map: dict[str | int, int]) -> None:
        r"""Initialise the CL class mapping transform object from the CL class map of a task.

        **Args:**
        - **cl_class_map** (`dict[str | int, int]`): the CL class map for a task.
        """
        self.cl_class_map = cl_class_map

    def __call__(self, target: torch.Tensor) -> torch.Tensor:
        r"""The CL class mapping transform to dataset labels. It is defined as a callable object like a PyTorch Transform.

        **Args:**
        - **target** (`Tensor`): the target tensor.

        **Returns:**
        - **transformed_target** (`Tensor`): the transformed target tensor.
        """

        return self.cl_class_map[target]


class Permute:
    r"""Permutation operation to image. Used to construct permuted CL dataset.

    Used as a PyTorch Dataset Transform.
    """

    def __init__(
        self,
        img_size: torch.Size,
        mode: str = "first_channel_only",
        seed: int | None = None,
    ) -> None:
        r"""Initialise the Permute transform object. The permutation order is constructed in the initialisation to save runtime.

        **Args:**
        - **img_size** (`torch.Size`): the size of the image to be permuted.
        - **mode** (`str`): the mode of permutation, shouble be one of the following:
            - 'all': permute all pixels.
            - 'by_channel': permute channel by channel separately. All channels are applied the same permutation order.
            - 'first_channel_only': permute only the first channel.
        - **seed** (`int` or `None`): seed for permutation operation. If None, the permutation will use a default seed from PyTorch generator.
        """
        self.mode = mode
        r"""Store the mode of permutation."""

        # get generator for permutation
        torch_generator = torch.Generator()
        if seed:
            torch_generator.manual_seed(seed)

        # calculate the number of pixels from the image size
        if self.mode == "all":
            num_pixels = img_size[0] * img_size[1] * img_size[2]
        elif self.mode == "by_channel" or "first_channel_only":
            num_pixels = img_size[1] * img_size[2]

        self.permute: torch.Tensor = torch.randperm(
            num_pixels, generator=torch_generator
        )
        r"""The permutation order, a `Tensor` permuted from [1,2, ..., `num_pixels`] with the given seed. It is the core element of permutation operation."""

    def __call__(self, img: torch.Tensor) -> torch.Tensor:
        r"""The permutation operation to image is defined as a callable object like a PyTorch Transform.

        **Args:**
        - **img** (`Tensor`): image to be permuted. Must match the size of `img_size` in the initialisation.

        **Returns:**
        - **img_permuted** (`Tensor`): the permuted image.
        """

        if self.mode == "all":

            img_flat = img.view(
                -1
            )  # flatten the whole image to 1d so that it can be applied 1d permuted order
            img_flat_permuted = img_flat[self.permute]  # conduct permutation operation
            img_permuted = img_flat_permuted.view(
                img.size()
            )  # return to the original image shape
            return img_permuted

        if self.mode == "by_channel":

            permuted_channels = []
            for i in range(img.size(0)):
                # act on every channel
                channel_flat = img[i].view(
                    -1
                )  # flatten the channel to 1d so that it can be applied 1d permuted order
                channel_flat_permuted = channel_flat[
                    self.permute
                ]  # conduct permutation operation
                channel_permuted = channel_flat_permuted.view(
                    img[0].size()
                )  # return to the original channel shape
                permuted_channels.append(channel_permuted)
            img_permuted = torch.stack(
                permuted_channels
            )  # stack the permuted channels to restore the image
            return img_permuted

        if self.mode == "first_channel_only":

            first_channel_flat = img[0].view(
                -1
            )  # flatten the first channel to 1d so that it can be applied 1d permuted order
            first_channel_flat_permuted = first_channel_flat[
                self.permute
            ]  # conduct permutation operation
            first_channel_permuted = first_channel_flat_permuted.view(
                img[0].size()
            )  # return to the original channel shape

            img_permuted = img.clone()
            img_permuted[0] = first_channel_permuted

            return img_permuted



r"""
The submodule in `cl_algorithms` for Fix algorithm.
"""

__all__ = ["Fix"]

import logging
from typing import Any

from torch import Tensor

from clarena.backbones import CLBackbone
from clarena.cl_algorithms import Finetuning
from clarena.cl_heads import HeadsCIL, HeadsTIL

# always get logger for built-in logging in each module
pylogger = logging.getLogger(__name__)


class Fix(Finetuning):
    r"""Fix algorithm.

    It is another naive way for task-incremental learning aside from Finetuning. It serves as kind of toy algorithm when discussing stability-plasticity dilemma in continual learning. It simply fixes the backbone forever after training first task.

    We implement Fix as a subclass of Finetuning algorithm, as Fix has the same `forward()`, `validation_step()` and `test_step()` method as `Finetuning` class.
    """

    def __init__(
        self,
        backbone: CLBackbone,
        heads: HeadsTIL | HeadsCIL,
    ) -> None:
        r"""Initialise the Fix algorithm with the network. It has no additional hyperparamaters.

        **Args:**
        - **backbone** (`CLBackbone`): backbone network.
        - **heads** (`HeadsTIL` | `HeadsCIL`): output heads.
        """
        Finetuning.__init__(self, backbone=backbone, heads=heads)

    def training_step(self, batch: Any) -> dict[str, Tensor]:
        """Training step for current task `self.task_id`.

        **Args:**
        - **batch** (`Any`): a batch of training data.

        **Returns:**
        - **outputs** (`dict[str, Tensor]`): a dictionary contains loss and other metrics from this training step. Key (`str`) is the metrics name, value (`Tensor`) is the metrics. Must include the key 'loss' which is total loss in the case of automatic optimization, according to PyTorch Lightning docs.
        """
        x, y = batch

        if self.task_id != 1:
            # Fix the backbone after training the first task
            for param in self.backbone.parameters():
                param.requires_grad = False

        # classification loss
        logits, hidden_features = self.forward(x, stage="train", task_id=self.task_id)
        loss_cls = self.criterion(logits, y)

        # total loss
        loss = loss_cls

        # accuracy of the batch
        acc = (logits.argmax(dim=1) == y).float().mean()

        return {
            "loss": loss,  # Return loss is essential for training step, or backpropagation will fail
            "loss_cls": loss_cls,
            "acc": acc,
            "hidden_features": hidden_features,
        }



r"""
The submodule in `backbones` for MLP backbone network.
"""

__all__ = ["MLP", "HATMaskMLP"]

from torch import Tensor, nn

from clarena.backbones import CLBackbone, HATMaskBackbone


class MLP(CLBackbone):
    """**Multi-layer perceptron (MLP)** a.k.a. fully-connected network.

    MLP is an dense network architecture, which has several fully-connected layers, each followed by an activation function. The last layer connects to the CL output heads.
    """

    def __init__(
        self,
        input_dim: int,
        hidden_dims: list[int],
        output_dim: int,
        activation_layer: nn.Module | None = nn.ReLU,
        batch_normalisation: bool = False,
        bias: bool = True,
        dropout: float | None = None,
    ) -> None:
        r"""Construct and initialise the MLP backbone network.

        **Args:**
        - **input_dim** (`int`): the input dimension. Any data need to be flattened before going in MLP. Note that it is not required in convolutional networks.
        - **hidden_dims** (`list[int]`): list of hidden layer dimensions. It can be empty list which means single-layer MLP, and it can be as many layers as you want. Note that it doesn't include the last dimension which we take as output dimension.
        - **output_dim** (`int`): the output dimension which connects to CL output heads.
        - **activation_layer** (`nn.Module` | `None`): activation function of each layer (if not `None`), if `None` this layer won't be used. Default `nn.ReLU`.
        - **batch_normalisation** (`bool`): whether to use batch normalisation after the fully-connected layers. Default `False`.
        - **bias** (`bool`): whether to use bias in the linear layer. Default `True`.
        - **dropout** (`float` | `None`): the probability for the dropout layer, if `None` this layer won't be used. Default `None`.
        """
        CLBackbone.__init__(self, output_dim=output_dim)

        self.num_fc_layers: int = len(hidden_dims) + 1
        r"""Store the number of fully-connected layers in the MLP backbone network, which helps form the loops in constructing layers and forward pass."""
        self.batch_normalisation: bool = batch_normalisation
        r"""Store whether to use batch normalisation after the fully-connected layers."""
        self.activation: bool = activation_layer is not None
        r"""Store whether to use activation function after the fully-connected layers."""
        self.dropout: bool = dropout is not None
        r"""Store whether to use dropout after the fully-connected layers."""

        self.fc: nn.ModuleList = nn.ModuleList()
        r"""The list of fully-connected (`nn.Linear`) layers. """
        if self.batch_normalisation:
            self.fc_bn: nn.ModuleList = nn.ModuleList()
            r"""The list of batch normalisation (`nn.BatchNorm1d`) layers after the fully-connected layers."""
        if self.activation:
            self.fc_activation: nn.ModuleList = nn.ModuleList()
            r"""The list of activation layers after the fully-connected layers. """
        if self.dropout:
            self.fc_dropout: nn.ModuleList = nn.ModuleList()
            r"""The list of dropout layers after the fully-connected layers. """

        # construct the weighted fully-connected layers and attached layers (batchnorm, activation, dropout, etc) in a loop
        for layer_idx in range(self.num_fc_layers):

            # the input and output dim of the current weighted layer
            layer_input_dim = (
                input_dim if layer_idx == 0 else hidden_dims[layer_idx - 1]
            )
            layer_output_dim = (
                hidden_dims[layer_idx] if layer_idx != len(hidden_dims) else output_dim
            )

            # construct the fully connected layer
            self.fc.append(
                nn.Linear(
                    in_features=layer_input_dim,
                    out_features=layer_output_dim,
                    bias=bias,
                )
            )

            # update the weighted layer names
            full_layer_name = f"fc/{layer_idx}"
            self.weighted_layer_names.append(full_layer_name)

            # construct the batch normalisation layer
            if self.batch_normalisation:
                self.fc_bn.append(nn.BatchNorm1d(num_features=(layer_output_dim)))

            # construct the activation layer
            if self.activation:
                self.fc_activation.append(activation_layer())

            # construct the dropout layer
            if self.dropout:
                self.fc_dropout.append(nn.Dropout(dropout))

    def forward(
        self, input: Tensor, stage: str = None, task_id: int | None = None
    ) -> tuple[Tensor, dict[str, Tensor]]:
        r"""The forward pass for data. It is the same for all tasks.

        **Args:**
        - **input** (`Tensor`): the input tensor from data.

        **Returns:**
        - **output_feature** (`Tensor`): the output feature tensor to be passed into heads. This is the main target of backpropagation.
        - **hidden_features** (`dict[str, Tensor]`): the hidden features (after activation) in each weighted layer. Key (`str`) is the weighted layer name, value (`Tensor`) is the hidden feature tensor. This is used for the continual learning algorithms that need to use the hidden features for various purposes.
        """
        batch_size = input.size(0)
        hidden_features = {}

        x = input.view(batch_size, -1)  # flatten before going through MLP

        for layer_idx, layer_name in enumerate(self.weighted_layer_names):
            x = self.fc[layer_idx](x)  # fully-connected layer first
            if self.batch_normalisation:
                x = self.fc_bn[layer_idx](x)  # batch normalisation second
            if self.activation:
                x = self.fc_activation[layer_idx](x)  # activation function third
            hidden_features[layer_name] = x  # store the hidden feature
            if self.dropout:
                x = self.fc_dropout[layer_idx](x)  # dropout last

        output_feature = x

        return output_feature, hidden_features


class HATMaskMLP(MLP, HATMaskBackbone):
    r"""HAT masked multi-Layer perceptron (MLP).

    [HAT (Hard Attention to the Task, 2018)](http://proceedings.mlr.press/v80/serra18a) is an architecture-based continual learning approach that uses learnable hard attention masks to select the task-specific parameters.

    MLP is a dense network architecture, which has several fully-connected layers, each followed by an activation function. The last layer connects to the CL output heads.

    Mask is applied to the units which are neurons in each fully-connected layer. The mask is generated from the unit-wise task embedding and gate function.
    """

    def __init__(
        self,
        input_dim: int,
        hidden_dims: list[int],
        output_dim: int,
        gate: str,
        activation_layer: nn.Module | None = nn.ReLU,
        bias: bool = True,
        dropout: float | None = None,
    ) -> None:
        r"""Construct and initialise the HAT masked MLP backbone network with task embedding. Note that batch normalisation is incompatible with HAT mechanism.

        **Args:**
        - **input_dim** (`int`): the input dimension. Any data need to be flattened before going in MLP.
        - **hidden_dims** (`list[int]`): list of hidden layer dimensions. It can be empty list which means single-layer MLP, and it can be as many layers as you want. Note that it doesn't include the last dimension which we take as output dimension.
        - **output_dim** (`int`): the output dimension which connects to CL output heads.
        - **gate** (`str`): the type of gate function turning the real value task embeddings into attention masks, should be one of the following:
            - `sigmoid`: the sigmoid function.
        - **activation_layer** (`nn.Module` | `None`): activation function of each layer (if not `None`), if `None` this layer won't be used. Default `nn.ReLU`.
        - **bias** (`bool`): whether to use bias in the linear layer. Default `True`.
        - **dropout** (`float` | `None`): the probability for the dropout layer, if `None` this layer won't be used. Default `None`.
        """
        # init from both inherited classes
        HATMaskBackbone.__init__(self, output_dim=output_dim, gate=gate)
        MLP.__init__(
            self,
            input_dim=input_dim,
            hidden_dims=hidden_dims,
            output_dim=output_dim,
            activation_layer=activation_layer,
            batch_normalisation=False,  # batch normalisation is incompatible with HAT mechanism
            bias=bias,
            dropout=dropout,
        )
        self.register_hat_mask_module_explicitly(
            gate=gate
        )  # register all `nn.Module`s for HATMaskBackbone explicitly because the second `__init__()` wipes out them inited by the first `__init__()`

        # construct the task embedding for each weighted layer
        for layer_idx in range(self.num_fc_layers):
            full_layer_name = f"fc/{layer_idx}"
            layer_output_dim = (
                hidden_dims[layer_idx] if layer_idx != len(hidden_dims) else output_dim
            )
            self.task_embedding_t[full_layer_name] = nn.Embedding(
                num_embeddings=1, embedding_dim=layer_output_dim
            )

    def forward(
        self,
        input: Tensor,
        stage: str,
        s_max: float | None = None,
        batch_idx: int | None = None,
        num_batches: int | None = None,
        test_mask: dict[str, Tensor] | None = None,
    ) -> tuple[Tensor, dict[str, Tensor], dict[str, Tensor]]:
        r"""The forward pass for data from task `task_id`. Task-specific mask for `task_id` are applied to the units which are neurons in each fully-connected layer.

        **Args:**
        - **input** (`Tensor`): the input tensor from data.
        - **stage** (`str`): the stage of the forward pass, should be one of the following:
            1. 'train': training stage.
            2. 'validation': validation stage.
            3. 'test': testing stage.
        - **s_max** (`float` | `None`): the maximum scaling factor in the gate function. Doesn't apply to testing stage. See chapter 2.4 "Hard Attention Training" in [HAT paper](http://proceedings.mlr.press/v80/serra18a).
        - **batch_idx** (`int` | `None`): the current batch index. Applies only to training stage. For other stages, it is default `None`.
        - **num_batches** (`int` | `None`): the total number of batches. Applies only to training stage. For other stages, it is default `None`.
        - **test_mask** (`dict[str, Tensor]` | `None`): the binary mask used for test. Applies only to testing stage. For other stages, it is default `None`.

        **Returns:**
        - **output_feature** (`Tensor`): the output feature tensor to be passed into heads. This is the main target of backpropagation.
        - **mask** (`dict[str, Tensor]`): the mask for the current task. Key (`str`) is layer name, value (`Tensor`) is the mask tensor. The mask tensor has size (number of units).
        - **hidden_features** (`dict[str, Tensor]`): the hidden features (after activation) in each weighted layer. Key (`str`) is the weighted layer name, value (`Tensor`) is the hidden feature tensor. This is used for the continual learning algorithms that need to use the hidden features for various purposes. Although HAT algorithm does not need this, it is still provided for API consistence for other HAT-based algorithms inherited this `forward()` method of `HAT` class.
        """
        batch_size = input.size(0)
        hidden_features = {}

        # get the mask for the current task from the task embedding in this stage
        mask = self.get_mask(
            stage=stage,
            s_max=s_max,
            batch_idx=batch_idx,
            num_batches=num_batches,
            test_mask=test_mask,
        )

        x = input.view(batch_size, -1)  # flatten before going through MLP

        for layer_idx, layer_name in enumerate(self.weighted_layer_names):
            x = self.fc[layer_idx](x)  # fully-connected layer first
            x = x * mask[f"fc/{layer_idx}"]  # apply the mask to the parameters second
            if self.activation:
                x = self.fc_activation[layer_idx](x)  # activation function third
            hidden_features[layer_name] = x  # store the hidden feature
            if self.dropout:
                x = self.fc_dropout[layer_idx](x)  # dropout last

        output_feature = x

        return output_feature, mask, hidden_features



r"""
The submodule in `callbacks` for `CLRichProgressBar`.
"""

__all__ = ["CLRichProgressBar"]

import logging

from lightning.pytorch.callbacks import RichProgressBar

# always get logger for built-in logging in each module
pylogger = logging.getLogger(__name__)


class CLRichProgressBar(RichProgressBar):
    r"""Customised `RichProgressBar` for continual learning."""

    def get_metrics(
        self, *args, **kwargs
    ) -> dict[str, int | str | float | dict[str, float]]:
        r"""Filter out the version number from the metrics displayed in the progress bar."""
        items = super().get_metrics(*args, **kwargs)
        items.pop("v_num", None)  # Remove the version number entry
        return items



"""

# Utilities

This submodule provides utilities that are used in CLArena, which includes:

- **Metrics**: define metrics particularly for CL.
- **Save**: for saving results to files.
- **Plot**: for plotting figures.


"""

from .cfg import preprocess_config
from .metrics import HATNetworkCapacity, MeanMetricBatch
from .plot import (
    plot_test_ave_acc_curve_from_csv,
    plot_test_ave_loss_cls_curve_from_csv,
    plot_hat_mask,
    plot_test_acc_matrix_from_csv,
    plot_test_loss_cls_matrix_from_csv,
)
from .save import update_test_acc_to_csv, update_test_loss_cls_to_csv
from .transforms import min_max_normalise

__all__ = ["cfg", "metrics", "save", "plot", "transforms"]



r"""
The submodule in `callbacks` for `SaveFirstBatchImagesCallback`.
"""

__all__ = ["SaveFirstBatchImagesCallback"]


import logging
import os

import torch
import torchvision
from lightning import Callback, Trainer

from clarena.cl_algorithms import CLAlgorithm

# always get logger for built-in logging in each module
pylogger = logging.getLogger(__name__)


class SaveFirstBatchImagesCallback(Callback):
    r"""Saves images and labels into files in the first batch of training data."""

    def __init__(
        self,
        save_dir: str,
        img_prefix: str = "sample",
        labels_filename: str = "labels.txt",
    ) -> None:
        r"""Initialise the Image Show Callback.

        **Args:**
        - **save_dir** (`str`): the directory to save images and labels as documents. Better inside the output directory.
        - **img_prefix** (`str`): the prefix for image files.
        - **labels_filename** (`str`): the filename for the labels file.
        """
        Callback.__init__(self)

        self.save_dir: str = save_dir
        r"""Store the directory to save images and labels as documents."""
        self.img_prefix: str = img_prefix
        r"""Store the prefix for image files."""
        self.labels_filename: str = labels_filename
        r"""Store the filename for the labels file."""

        self.called: bool = False
        r"""Flag to avoid calling the callback multiple times."""

    def on_train_start(self, trainer: Trainer, pl_module: CLAlgorithm) -> None:
        r"""Save images and labels into files in the first batch of training data at the beginning of the training of the task."""
        if self.called:
            return  # flag to avoid calling the callback multiple times

        dataloader = trainer.train_dataloader
        image_batch, label_batch = next(iter(dataloader))  # the first batch
        image_samples = list(torch.unbind(image_batch, dim=0))
        label_samples = list(torch.unbind(label_batch, dim=0))

        # save images and labels as documents
        save_dir_task = os.path.join(self.save_dir, f"task_{pl_module.task_id}")
        os.makedirs(save_dir_task, exist_ok=True)
        labels_file = open(
            os.path.join(save_dir_task, self.labels_filename),
            "w",
            encoding="utf-8",
        )
        for i, (image, label) in enumerate(zip(image_samples, label_samples)):
            torchvision.utils.save_image(
                image, os.path.join(save_dir_task, f"{self.img_prefix}_{i}.png")
            )
            labels_file.write(f"{self.img_prefix}_{i}.png: {label}\n")
        labels_file.close()

        self.called = True  # flag to avoid calling the callback multiple times



r"""
The submodule in `cl_algorithms` for [HAT (Hard Attention to the Task) algorithm](http://proceedings.mlr.press/v80/serra18a).
"""

__all__ = ["HAT"]

import logging
from typing import Any

import torch
from torch import Tensor
from torch.utils.data import DataLoader

from clarena.backbones import HATMaskBackbone
from clarena.cl_algorithms import CLAlgorithm
from clarena.cl_algorithms.regularisers import HATMaskSparsityReg
from clarena.cl_heads import HeadsCIL, HeadsTIL
from clarena.utils import HATNetworkCapacity

# always get logger for built-in logging in each module
pylogger = logging.getLogger(__name__)


class HAT(CLAlgorithm):
    r"""HAT (Hard Attention to the Task) algorithm.

    [HAT (Hard Attention to the Task, 2018)](http://proceedings.mlr.press/v80/serra18a) is an architecture-based continual learning approach that uses learnable hard attention masks to select the task-specific parameters.

    """

    def __init__(
        self,
        backbone: HATMaskBackbone,
        heads: HeadsTIL | HeadsCIL,
        adjustment_mode: str,
        s_max: float,
        clamp_threshold: float,
        mask_sparsity_reg_factor: float,
        mask_sparsity_reg_mode: str = "original",
        task_embedding_init_mode: str = "N01",
        alpha: float | None = None,
    ) -> None:
        r"""Initialise the HAT algorithm with the network.

        **Args:**
        - **backbone** (`HATMaskBackbone`): must be a backbone network with HAT mask mechanism.
        - **heads** (`HeadsTIL` | `HeadsCIL`): output heads.
        - **adjustment_mode** (`str`): the strategy of adjustment i.e. the mode of gradient clipping, should be one of the following:
            1. 'hat': set the gradients of parameters linking to masked units to zero. This is the way that HAT does, which fixes the part of network for previous tasks completely. See equation (2) in chapter 2.3 "Network Training" in [HAT paper](http://proceedings.mlr.press/v80/serra18a).
            2. 'hat_random': set the gradients of parameters linking to masked units to random 0-1 values. See the "Baselines" section in chapter 4.1 in [AdaHAT paper](https://link.springer.com/chapter/10.1007/978-3-031-70352-2_9).
            3. 'hat_const_alpha': set the gradients of parameters linking to masked units to a constant value of `alpha`. See the "Baselines" section in chapter 4.1 in [AdaHAT paper](https://link.springer.com/chapter/10.1007/978-3-031-70352-2_9).
            4. 'hat_const_1': set the gradients of parameters linking to masked units to a constant value of 1, which means no gradient constraint on any parameter at all. See the "Baselines" section in chapter 4.1 in [AdaHAT paper](https://link.springer.com/chapter/10.1007/978-3-031-70352-2_9).
        - **s_max** (`float`): hyperparameter, the maximum scaling factor in the gate function. See chapter 2.4 "Hard Attention Training" in [HAT paper](http://proceedings.mlr.press/v80/serra18a).
        - **clamp_threshold** (`float`): the threshold for task embedding gradient compensation. See chapter 2.5 "Embedding Gradient Compensation" in [HAT paper](http://proceedings.mlr.press/v80/serra18a).
        - **mask_sparsity_reg_factor** (`float`): hyperparameter, the regularisation factor for mask sparsity.
        - **mask_sparsity_reg_mode** (`str`): the mode of mask sparsity regularisation, should be one of the following:
            1. 'original' (default): the original mask sparsity regularisation in HAT paper.
            2. 'cross': the cross version mask sparsity regularisation.
        - **task_embedding_init_mode** (`str`): the initialisation mode for task embeddings, should be one of the following:
            1. 'N01' (default): standard normal distribution $N(0, 1)$.
            2. 'U-11': uniform distribution $U(-1, 1)$.
            3. 'U01': uniform distribution $U(0, 1)$.
            4. 'U-10': uniform distribution $U(-1, 0)$.
            5. 'last': inherit task embedding from last task.
        - **alpha** (`float` | `None`): the `alpha` in the 'HAT-const-alpha' mode. See the "Baselines" section in chapter 4.1 in [AdaHAT paper](https://link.springer.com/chapter/10.1007/978-3-031-70352-2_9). It applies only when adjustment_mode is 'hat_const_alpha'.
        """
        CLAlgorithm.__init__(self, backbone=backbone, heads=heads)

        self.adjustment_mode = adjustment_mode
        r"""Store the adjustment mode for gradient clipping."""
        self.s_max = s_max
        r"""Store s_max. """
        self.clamp_threshold = clamp_threshold
        r"""Store the clamp threshold for task embedding gradient compensation."""
        self.mask_sparsity_reg_factor = mask_sparsity_reg_factor
        r"""Store the mask sparsity regularisation factor."""
        self.mask_sparsity_reg_mode = mask_sparsity_reg_mode
        r"""Store the mask sparsity regularisation mode."""
        self.mark_sparsity_reg = HATMaskSparsityReg(
            factor=mask_sparsity_reg_factor, mode=mask_sparsity_reg_mode
        )
        r"""Initialise and store the mask sparsity regulariser."""
        self.task_embedding_init_mode = task_embedding_init_mode
        r"""Store the task embedding initialisation mode."""
        self.alpha = alpha if adjustment_mode == "hat_const_alpha" else None
        r"""Store the alpha for `hat_const_alpha`."""
        self.epsilon = None
        r"""HAT doesn't use the epsilon for `hat_const_alpha`. We still set it here to be consistent with the `epsilon` in `clip_grad_by_adjustment()` method in `HATMaskBackbone`."""

        self.masks: dict[str, dict[str, Tensor]] = {}
        r"""Store the binary attention mask of each previous task gated from the task embedding. Keys are task IDs (string type) and values are the corresponding mask. Each mask is a dict where keys are layer names and values are the binary mask tensor for the layer. The mask tensor has size (number of units). """

        self.cumulative_mask_for_previous_tasks: dict[str, Tensor] = {}
        r"""Store the cumulative binary attention mask $\mathrm{M}^{<t}$ of previous tasks $1,\cdots, t-1$, gated from the task embedding. Keys are task IDs and values are the corresponding cumulative mask. Each cumulative mask is a dict where keys are layer names and values are the binary mask tensor for the layer. The mask tensor has size (number of units). """

        # set manual optimisation
        self.automatic_optimization = False

        HAT.sanity_check(self)

    def sanity_check(self) -> None:
        r"""Check the sanity of the arguments.

        **Raises:**
        - **ValueError**: when backbone is not designed for HAT, or the `mask_sparsity_reg_mode` or `task_embedding_init_mode` is not one of the valid options. Also, if `alpha` is not given when `adjustment_mode` is 'hat_const_alpha'.
        """
        if not isinstance(self.backbone, HATMaskBackbone):
            raise ValueError("The backbone should be an instance of HATMaskBackbone.")

        if self.mask_sparsity_reg_mode not in ["original", "cross"]:
            raise ValueError(
                "The mask_sparsity_reg_mode should be one of 'original', 'cross'."
            )
        if self.task_embedding_init_mode not in [
            "N01",
            "U01",
            "U-10",
            "masked",
            "unmasked",
        ]:
            raise ValueError(
                "The task_embedding_init_mode should be one of 'N01', 'U01', 'U-10', 'masked', 'unmasked'."
            )

        if self.adjustment_mode == "hat_const_alpha" and self.alpha is None:
            raise ValueError(
                "Alpha should be given when the adjustment_mode is 'hat_const_alpha'."
            )

    def on_train_start(self) -> None:
        r"""Initialise the task embedding before training the next task and initialise the cumulative mask at the beginning of first task."""

        self.backbone.initialise_task_embedding(mode=self.task_embedding_init_mode)

        # initialise the cumulative mask at the beginning of first task. This should not be called in `__init__()` method as the `self.device` is not available at that time.
        if self.task_id == 1:
            for layer_name in self.backbone.weighted_layer_names:
                layer = self.backbone.get_layer_by_name(
                    layer_name
                )  # get the layer by its name
                num_units = layer.weight.shape[0]

                self.cumulative_mask_for_previous_tasks[layer_name] = torch.zeros(
                    num_units
                ).to(
                    self.device
                )  # the cumulative mask $\mathrm{M}^{<t}$ is initialised as zeros mask ($t = 1$). See equation (2) in chapter 3 in [AdaHAT paper](https://link.springer.com/chapter/10.1007/978-3-031-70352-2_9), or equation (5) in chapter 2.6 "Promoting Low Capacity Usage" in [HAT paper](http://proceedings.mlr.press/v80/serra18a).

    def clip_grad_by_adjustment(
        self,
        **kwargs,
    ) -> Tensor:
        r"""Clip the gradients by the adjustment rate.

        Note that as the task embedding fully covers every layer in the backbone network, no parameters are left out of this system. This applies not only the parameters in between layers with task embedding, but also those before the first layer. We designed it seperately in the codes.

        Network capacity is measured along with this method. Network capacity is defined as the average adjustment rate over all parameters. See chapter 4.1 in [AdaHAT paper](https://link.springer.com/chapter/10.1007/978-3-031-70352-2_9).


        **Returns:**
        - **capacity** (`Tensor`): the calculated network capacity.
        """

        # initialise network capacity metric
        capacity = HATNetworkCapacity()

        # Calculate the adjustment rate for gradients of the parameters, both weights and biases (if exists)
        for layer_name in self.backbone.weighted_layer_names:

            layer = self.backbone.get_layer_by_name(
                layer_name
            )  # get the layer by its name

            # placeholder for the adjustment rate to avoid the error of using it before assignment
            adjustment_rate_weight = 1
            adjustment_rate_bias = 1

            weight_mask, bias_mask = self.backbone.get_layer_measure_parameter_wise(
                unit_wise_measure=self.cumulative_mask_for_previous_tasks,
                layer_name=layer_name,
                aggregation="min",
            )

            if self.adjustment_mode == "hat":
                adjustment_rate_weight = 1 - weight_mask
                adjustment_rate_bias = 1 - bias_mask

            elif self.adjustment_mode == "hat_random":
                adjustment_rate_weight = torch.rand_like(weight_mask) * weight_mask + (
                    1 - weight_mask
                )
                adjustment_rate_bias = torch.rand_like(bias_mask) * bias_mask + (
                    1 - bias_mask
                )

            elif self.adjustment_mode == "hat_const_alpha":
                adjustment_rate_weight = self.alpha * torch.ones_like(
                    weight_mask
                ) * weight_mask + (1 - weight_mask)
                adjustment_rate_bias = self.alpha * torch.ones_like(
                    bias_mask
                ) * bias_mask + (1 - bias_mask)

            elif self.adjustment_mode == "hat_const_1":
                adjustment_rate_weight = torch.ones_like(weight_mask) * weight_mask + (
                    1 - weight_mask
                )
                adjustment_rate_bias = torch.ones_like(bias_mask) * bias_mask + (
                    1 - bias_mask
                )

            # apply the adjustment rate to the gradients
            layer.weight.grad.data *= adjustment_rate_weight
            if layer.bias is not None:
                layer.bias.grad.data *= adjustment_rate_bias

            # update network capacity metric
            capacity.update(adjustment_rate_weight, adjustment_rate_bias)

        return capacity.compute()

    def compensate_task_embedding_gradients(
        self,
        batch_idx: int,
        num_batches: int,
    ) -> None:
        r"""Compensate the gradients of task embeddings during training. See chapter 2.5 "Embedding Gradient Compensation" in [HAT paper](http://proceedings.mlr.press/v80/serra18a).

        **Args:**
        - **batch_idx** (`int`): the current training batch index.
        - **num_batches** (`int`): the total number of training batches.
        """

        for te in self.backbone.task_embedding_t.values():
            anneal_scalar = 1 / self.s_max + (self.s_max - 1 / self.s_max) * (
                batch_idx - 1
            ) / (
                num_batches - 1
            )  # see equation (3) in chapter 2.4 "Hard Attention Training" in [HAT paper](http://proceedings.mlr.press/v80/serra18a)

            num = (
                torch.cosh(
                    torch.clamp(
                        anneal_scalar * te.weight.data,
                        -self.clamp_threshold,
                        self.clamp_threshold,
                    )
                )
                + 1
            )

            den = torch.cosh(te.weight.data) + 1

            compensation = self.s_max / anneal_scalar * num / den

            te.weight.grad.data *= compensation

    def forward(
        self,
        input: torch.Tensor,
        stage: str,
        batch_idx: int | None = None,
        num_batches: int | None = None,
        task_id: int | None = None,
    ) -> tuple[Tensor, dict[str, Tensor]]:
        r"""The forward pass for data from task `task_id`. Note that it is nothing to do with `forward()` method in `nn.Module`.

        **Args:**
        - **input** (`Tensor`): The input tensor from data.
        - **stage** (`str`): the stage of the forward pass, should be one of the following:
            1. 'train': training stage.
            2. 'validation': validation stage.
            3. 'test': testing stage.
        - **batch_idx** (`int` | `None`): the current batch index. Applies only to training stage. For other stages, it is default `None`.
        - **num_batches** (`int` | `None`): the total number of batches. Applies only to training stage. For other stages, it is default `None`.
        - **task_id** (`int`| `None`): the task ID where the data are from. If the stage is 'train' or 'validation', it should be the current task `self.task_id`. If stage is 'test', it could be from any seen task. In TIL, the task IDs of test data are provided thus this argument can be used. HAT algorithm works only for TIL.

        **Returns:**
        - **logits** (`Tensor`): the output logits tensor.
        - **mask** (`dict[str, Tensor]`): the mask for the current task. Key (`str`) is layer name, value (`Tensor`) is the mask tensor. The mask tensor has size (number of units).
        - **hidden_features** (`dict[str, Tensor]`): the hidden features (after activation) in each weighted layer. Key (`str`) is the weighted layer name, value (`Tensor`) is the hidden feature tensor. This is used for the continual learning algorithms that need to use the hidden features for various purposes. Although HAT algorithm does not need this, it is still provided for API consistence for other HAT-based algorithms inherited this `forward()` method of `HAT` class.
        """
        feature, mask, hidden_features = self.backbone(
            input,
            stage=stage,
            s_max=self.s_max if stage == "train" or stage == "validation" else None,
            batch_idx=batch_idx if stage == "train" else None,
            num_batches=num_batches if stage == "train" else None,
            test_mask=self.masks[f"{task_id}"] if stage == "test" else None,
        )
        logits = self.heads(feature, task_id)

        return logits, mask, hidden_features

    def training_step(self, batch: Any, batch_idx: int) -> dict[str, Tensor]:
        r"""Training step for current task `self.task_id`.

        **Args:**
        - **batch** (`Any`): a batch of training data.
        - **batch_idx** (`int`): the index of the batch. Used for calculating annealed scalar in HAT. See chapter 2.4 "Hard Attention Training" in [HAT paper](http://proceedings.mlr.press/v80/serra18a).

        **Returns:**
        - **outputs** (`dict[str, Tensor]`): a dictionary contains loss and other metrics from this training step. Key (`str`) is the metrics name, value (`Tensor`) is the metrics. Must include the key 'loss' which is total loss in the case of automatic optimization, according to PyTorch Lightning docs. For HAT, it includes 'mask' and 'capacity' for logging.
        """
        x, y = batch

        # zero the gradients before forward pass in manual optimisation mode
        opt = self.optimizers()
        opt.zero_grad()

        # classification loss
        num_batches = self.trainer.num_training_batches
        logits, mask, hidden_features = self.forward(
            x,
            stage="train",
            batch_idx=batch_idx,
            num_batches=num_batches,
            task_id=self.task_id,
        )
        loss_cls = self.criterion(logits, y)

        # regularisation loss. See chapter 2.6 "Promoting Low Capacity Usage" in [HAT paper](http://proceedings.mlr.press/v80/serra18a).
        loss_reg, network_sparsity = self.mark_sparsity_reg(
            mask, self.cumulative_mask_for_previous_tasks
        )

        # total loss
        loss = loss_cls + loss_reg

        # backward step (manually)
        self.manual_backward(loss)  # calculate the gradients
        # HAT hard clip gradients by the cumulative masks. See equation (2) inchapter 2.3 "Network Training" in [HAT paper](http://proceedings.mlr.press/v80/serra18a). Network capacity is calculated along with this process. Network capacity is defined as the average adjustment rate over all paramaters. See chapter 4.1 in [AdaHAT paper](https://link.springer.com/chapter/10.1007/978-3-031-70352-2_9).
        capacity = self.clip_grad_by_adjustment(
            network_sparsity=network_sparsity,  # pass a keyword argument network sparsity here to make it compatible with AdaHAT. AdaHAT inherits this `training_step()` method.
        )
        # compensate the gradients of task embedding. See chapter 2.5 "Embedding Gradient Compensation" in [HAT paper](http://proceedings.mlr.press/v80/serra18a).
        self.compensate_task_embedding_gradients(
            batch_idx=batch_idx,
            num_batches=num_batches,
        )
        # update parameters with the modified gradients
        opt.step()

        # accuracy of the batch
        acc = (logits.argmax(dim=1) == y).float().mean()

        return {
            "loss": loss,  # Return loss is essential for training step, or backpropagation will fail
            "loss_cls": loss_cls,
            "loss_reg": loss_reg,
            "acc": acc,
            "hidden_features": hidden_features,
            "mask": mask,  # Return other metrics for lightning loggers callback to handle at `on_train_batch_end()`
            "capacity": capacity,
        }

    def on_train_end(self) -> None:
        r"""Store the mask and update cumulative mask after training the task."""

        # store the mask for the current task
        mask_t = {
            layer_name: self.backbone.gate_fn(
                self.backbone.task_embedding_t[layer_name].weight * self.s_max
            )
            .squeeze()
            .detach()
            for layer_name in self.backbone.weighted_layer_names
        }

        self.masks[f"{self.task_id}"] = mask_t

        # update the cumulative and summative masks
        self.cumulative_mask_for_previous_tasks = {
            layer_name: torch.max(
                self.cumulative_mask_for_previous_tasks[layer_name], mask_t[layer_name]
            )
            for layer_name in self.backbone.weighted_layer_names
        }

    def validation_step(self, batch: Any) -> dict[str, Tensor]:
        r"""Validation step for current task `self.task_id`.

        **Args:**
        - **batch** (`Any`): a batch of validation data.

        **Returns:**
        - **outputs** (`dict[str, Tensor]`): a dictionary contains loss and other metrics from this validation step. Key (`str`) is the metrics name, value (`Tensor`) is the metrics.
        """
        x, y = batch
        logits, mask, hidden_features = self.forward(
            x, stage="validation", task_id=self.task_id
        )
        loss_cls = self.criterion(logits, y)
        acc = (logits.argmax(dim=1) == y).float().mean()

        return {
            "loss_cls": loss_cls,
            "acc": acc,  # Return metrics for lightning loggers callback to handle at `on_validation_batch_end()`
        }

    def test_step(
        self, batch: DataLoader, batch_idx: int, dataloader_idx: int = 0
    ) -> dict[str, Tensor]:
        r"""Test step for current task `self.task_id`, which tests for all seen tasks indexed by `dataloader_idx`.

        **Args:**
        - **batch** (`Any`): a batch of test data.
        - **dataloader_idx** (`int`): the task ID of seen tasks to be tested. A default value of 0 is given otherwise the LightningModule will raise a `RuntimeError`.

        **Returns:**
        - **outputs** (`dict[str, Tensor]`): a dictionary contains loss and other metrics from this test step. Key (`str`) is the metrics name, value (`Tensor`) is the metrics.
        """
        test_task_id = dataloader_idx + 1

        x, y = batch
        logits, mask, hidden_features = self.forward(
            x,
            stage="test",
            task_id=test_task_id,
        )  # use the corresponding head and mask to test (instead of the current task `self.task_id`)
        loss_cls = self.criterion(logits, y)
        acc = (logits.argmax(dim=1) == y).float().mean()

        return {
            "loss_cls": loss_cls,
            "acc": acc,  # Return metrics for lightning loggers callback to handle at `on_test_batch_end()`
        }



r"""

# Backbone Networks for Continual Learning

This submodule provides the **backbone neural network architectures for continual learning**. 

Please note that this is an API documentation. Please refer to the main documentation pages for more information about the backbone networks and how to 
configure and implement them:

- [**Configure Backbone Network**](https://pengxiang-wang.com/projects/continual-learning-arena/docs/configure-your-experiment/backbone-network)
- [**Implement Your CL Backbone Class**](https://pengxiang-wang.com/projects/continual-learning-arena/docs/implement-your-CL-modules/backbone-network)



The backbones are implemented as subclasses of `CLBackbone` classes, which are the base class for all continual learning backbones in CLArena.

- `CLBackbone`: The base class for continual learning backbones.
- `HATMaskBackbone`: The base class for backbones used in [HAT (Hard Attention to the Task) algorithm](http://proceedings.mlr.press/v80/serra18a). A child class of `CLBackbone`.


"""

from .base import CLBackbone, HATMaskBackbone
from .mlp import MLP, HATMaskMLP
from .resnet import (
    HATMaskResNet18,
    HATMaskResNet34,
    HATMaskResNet50,
    HATMaskResNet101,
    HATMaskResNet152,
    ResNet18,
    ResNet34,
    ResNet50,
    ResNet101,
    ResNet152,
)

__all__ = ["CLBackbone", "HATMaskBackbone", "mlp", "resnet"]
